from amuse.test import amusetest
from amuse.units import units, nbody_system, constants
from amuse.ext.galactic_potentials import NFW_profile, MiyamotoNagai_profile, Plummer_profile, \
    PowerLawCutoff_profile, MWpotentialBovy2015, scipy_imported

import numpy


class TestNFWProfile(amusetest.TestCase):
    def test1(self):
        """
        See the textbook by L.Aguilar, chapter 3:
        ftp://ftp.crya.unam.mx/pub/luisfr/laguilar/GH05_Aguilar.pdf
        """
        rho0 = 1.e3 | units.MSun/units.parsec**3
        rs = 6.6 | units.kpc
        nfw_profile = NFW_profile(rho0, rs)
        m0 = 4.*numpy.pi*rho0*rs**3
        phi0 = -4.*numpy.pi*constants.G*rho0*rs**2
        ar0 = -phi0/rs

        # relative mass enclosed at rs: m_at_rs / m0 ~ 0.193147
        m_at_rs = nfw_profile.enclosed_mass(rs)
        self.assertAlmostEqual(m_at_rs/m0, 0.193147, 5)

        # relative mass enclosed at 5.3054*rs ~ m0
        m_eq_one = nfw_profile.enclosed_mass(5.3054*rs,)
        self.assertAlmostEqual(m_eq_one/m0, 1.0, 5)

        # relative mass density, rho(rs) ~ rho0/4
        rho_rs = nfw_profile.mass_density(rs)
        self.assertAlmostEqual(rho0/rho_rs, 4.0, 5)

        # relative gravitational potential at (r/rs)->0 is approaching 1
        phi_at_0 = nfw_profile.get_potential_at_point(0. | units.m, 0. | units.m, 0. | units.m, 1.e-10 | units.kpc)
        self.assertAlmostEqual(phi_at_0/phi0, 1.0, 4)

        # relative force at (r/rs)->0 is approaching -1/2
        ar_at_0 = nfw_profile.radial_force(1.e-5 | units.kpc)
        self.assertAlmostEqual(ar_at_0/ar0, -0.5, 4)

        # relative force at (r/rs)->inf is approaching 0
        ar_at_inf = nfw_profile.radial_force(1.e10 | units.kpc)
        self.assertAlmostEqual(ar_at_inf/ar0, 0.0, 5)

        # relative circular velocity has maximum at r/r_s=2.16258
        vc_eq_max = nfw_profile.circular_velocity(2.16258*rs)
        vc_at_r_lt_max = nfw_profile.circular_velocity(2.16248*rs)
        vc_at_r_gt_max = nfw_profile.circular_velocity(2.16268*rs)
        self.assertTrue(vc_at_r_lt_max < vc_eq_max)
        self.assertTrue(vc_at_r_gt_max < vc_eq_max)


class TestPlummerProfile(amusetest.TestCase):
    def test1(self):
        mass = 6.e6 | units.MSun
        a = 6. | units.parsec
        plummer_profile = Plummer_profile(mass, a)
        rho0 = mass/(4./3.*numpy.pi*a**3)
        phi0 = -constants.G*mass/a

        # enclosed mass at R>>a is total mass
        m_tot = plummer_profile.enclosed_mass(a*1.e5)
        self.assertAlmostEqual(m_tot/mass, 1.0, 5)

        # mass density at the center
        rho_cen = plummer_profile.mass_density(0. | units.m)
        self.assertAlmostEqual(rho_cen/rho0, 1.0, 5)

        # potential at r=a is phi0/sqrt(2)
        phi_at_a = plummer_profile.get_potential_at_point(0. | units.m, 0. | units.m, 0. | units.m, a)
        self.assertAlmostEqual(phi_at_a/phi0*numpy.sqrt(2.), 1.0, 5)


class TestMiyamotoNagaiProfile(amusetest.TestCase):
    def test1(self):
        mass = 6.6e10 | units.MSun
        a = 3.33 | units.kpc
        b = 0.666 | units.kpc
        profile = MiyamotoNagai_profile(mass, a, b)

        r_force = profile.force_R(0.1*a, 0.2*b, 1.2*a).in_(units.parsec/units.Myr**2)
        z_force = profile.force_z(0.1*a, 0.2*b, 1.2*a).in_(units.parsec/units.Myr**2)
        potential = profile.get_potential_at_point(0. | units.m, a*0.1, a*5., b*0.5).in_(units.kpc**2/units.Myr**2)
        ax, ay, az = profile.get_gravity_at_point(0. | units.m, a*0.1, a*5., b*0.5)
        density = profile.mass_density(0.1*a, -0.5*b, 1.2*a).in_(units.MSun/units.kpc**3)
        vc = profile.circular_velocity_at_z0(1.0*a).in_(units.kms)
        m_r = profile.equivalent_enclosed_mass_in_plane(100.*a).in_(units.MSun)

        self.assertAlmostEqual(r_force, -0.263920797645 | units.parsec/units.Myr**2, 12)
        self.assertAlmostEqual(z_force, -5.35763387945 | units.parsec/units.Myr**2, 11)
        self.assertAlmostEqual(potential, -0.017321166943 | units.kpc**2/units.Myr**2, 12)
        self.assertAlmostEqual(ax, -0.0196231550925 | units.parsec/units.Myr**2, 12)
        self.assertAlmostEqual(ay, -0.981157754625 | units.parsec/units.Myr**2, 12)
        self.assertAlmostEqual(az, -0.107380572532 | units.parsec/units.Myr**2, 12)
        self.assertAlmostEqual(density, 1336672.32264 | units.MSun/units.kpc**3, 5)
        self.assertAlmostEqual(vc, 149.569512197 | units.kms, 9)
        self.assertAlmostEqual(m_r, 65.9857465656 | 1.e9*units.MSun, 9)

    def test2(self):
        mass = 1.984e4 | units.MSun
        a = 0.0 | units.parsec
        b = 6.66 | units.parsec
        nm_profile = MiyamotoNagai_profile(mass, a, b)
        plummer_profile = Plummer_profile(mass, b)

        pot_nm = nm_profile.get_potential_at_point(0. | units.m, b*0.1, b*5., b*0.2)
        pot_p = plummer_profile.get_potential_at_point(0. | units.m, b*0.1, b*5., b*0.2)
        self.assertEqual(pot_nm, pot_p)

        ax_nm, ay_nm, az_nm = nm_profile.get_gravity_at_point(0. | units.m, b*0.1, b*5., b*0.1)
        ax_p, ay_p, az_p = plummer_profile.get_gravity_at_point(0. | units.m, b*0.1, b*5., b*0.1)
        print(ax_nm.in_(units.parsec/units.Myr**2), ax_p.in_(units.parsec/units.Myr**2))
        self.assertAlmostEqual(ax_nm.in_(units.parsec/units.Myr**2), ax_p.in_(units.parsec/units.Myr**2), 12)
        self.assertAlmostEqual(ay_nm.in_(units.parsec/units.Myr**2), ay_p.in_(units.parsec/units.Myr**2), 12)
        self.assertAlmostEqual(az_nm.in_(units.parsec/units.Myr**2), az_p.in_(units.parsec/units.Myr**2), 12)

        rho_nm = nm_profile.mass_density(b*0., b*0., b*6.6)
        rho_p = plummer_profile.mass_density(b*6.6)
        self.assertEqual(rho_nm, rho_p)


class TestPowerLawCutoff_profile(amusetest.TestCase):
    def test1(self):
        rho0 = 12. | units.MSun/units.parsec**3
        r0 = 1.6 | units.parsec
        alpha = 1.6
        rc = 0.66 | units.kpc
        power_law = PowerLawCutoff_profile(rho0, r0, alpha, rc)

        r_force = power_law.radial_force(0.1*r0).in_(units.parsec/units.Myr**2)
        potential = power_law.get_potential_at_point(0. | units.m, r0*0.1, r0*5., r0*0.5).in_(units.kpc**2/units.Myr**2)
        ax, ay, az = power_law.get_gravity_at_point(0. | units.m, r0*0.1, r0*5., r0*0.5)
        density = power_law.mass_density(6.6*r0).in_(units.MSun/units.parsec**3)
        vc = power_law.circular_velocity(r0).in_(units.kms)
        m_r = power_law.enclosed_mass(100.*r0).in_(units.MSun)

        self.assertAlmostEqual(r_force, -3.08704194743 | units.parsec/units.Myr**2, 10)
        self.assertAlmostEqual(potential, -5.91677122595e-06 | units.kpc**2/units.Myr**2, 10)
        self.assertAlmostEqual(ax, -0.00585557189412 | units.parsec/units.Myr**2, 10)
        self.assertAlmostEqual(ay, -0.292778594706 | units.parsec/units.Myr**2, 10)
        self.assertAlmostEqual(az, -0.0292778594706 | units.parsec/units.Myr**2, 10)
        self.assertAlmostEqual(density, 0.585867989506 | units.MSun/units.parsec**3, 10)
        self.assertAlmostEqual(vc, 1.0891472277 | units.kms, 10)
        self.assertAlmostEqual(m_r, 2.71756907682 | 1.e5*units.MSun, 9)

    def test2(self):
        rho0 = 12. | units.MSun/units.parsec**3
        r0 = 1.6 | units.parsec
        alpha = 1.6
        rc = 0.66 | units.kpc
        power_law = PowerLawCutoff_profile(rho0, r0, alpha, rc)

        r = 2*rc*numpy.arange(100001)/100000.

        pot = power_law.get_potential_at_point(0. | units.kpc, r, 0.*r, 0*r)
        rho = power_law.mass_density(r)

        d = r[1]-r[0]
        rpot = r*pot
        lapl = (rpot[2:]-2*rpot[1:-1]+rpot[:-2])/d**2/r[1:-1]
        rcheck = r[1:-1]

        rho1 = (-lapl/(4*units.pi*constants.G))
        rho2 = rho[1:-1]
        d = ((rho1[1:]-rho2[1:])/rho2[1:])
        # seems to be ok:
        self.assertTrue(d.max() < 0.025)
        self.assertTrue(d.mean() < 1.e-6)

    def setUp(self):
        if not scipy_imported:
            self.skip("scipy not installed")


class TestMWpotentialBovy2015(amusetest.TestCase):
    def test1(self):
        """
        See Table 1 of Bovy 2015, http://adsabs.harvard.edu/abs/2015ApJS..216...29B
        """
        mw = MWpotentialBovy2015()
        r0 = 8. | units.kpc
        v0 = 220. | units.kms

        # total mass density at r=r0,z=0
        rho_r0_z0 = mw.mass_density(r0, 0. | units.m, 0. | units.m)
        self.assertAlmostEqual(rho_r0_z0, 0.10 | units.MSun/units.parsec**3, 2)

        # halo mass density at r0
        rho_halo_at_r0 = mw.halo.mass_density(r0)
        self.assertAlmostEqual(rho_halo_at_r0, 0.008 | units.MSun/units.parsec**3, 3)

        # mass enclosed in 60kpc
        mass_in_60 = mw.enclosed_mass(60. | units.kpc)
        print(mass_in_60.in_(units.MSun))
        self.assertAlmostEqual((mass_in_60/1.e11).in_(units.MSun), 4.08 | units.MSun, 2)

        # normalization factor for bulge
        fr_r0_v0 = v0**2 / r0
        fr_r0_bulge = -mw.bulge.radial_force(r0)
        self.assertAlmostEqual(fr_r0_bulge/fr_r0_v0, 0.05, 3)

        # normalization factor for disk
        fr_r0_disk = -mw.disk.force_R(r0, 0. | units.m, 0. | units.m)
        self.assertAlmostEqual(fr_r0_disk/fr_r0_v0, 0.6, 3)

        # normalization factor for halo
        fr_r0_halo = -mw.halo.radial_force(r0)
        print(fr_r0_halo/fr_r0_v0)
        self.assertAlmostEqual(fr_r0_halo/fr_r0_v0, 0.35, 3)

    def setUp(self):
        if not scipy_imported:
            self.skip("scipy not installed")



from amuse.test import amusetest

from amuse.community.sse.interface import SSE
from amuse.community.bhtree.interface import BHTree


class TestsForIssue123(amusetest.TestCase):
    def test1(self):  # doesn't trigger recursion error
        self.assertRaises(Exception, BHTree, name_of_the_worker="bogus", expected_message="__init__() got multiple values for keyword argument 'name_of_the_worker'")

    def test2(self):  # does
        self.assertRaises(Exception, SSE, name_of_the_worker="bogus", expected_message="__init__() got multiple values for keyword argument 'name_of_the_worker'")



import sys
import os
import numpy.random

from amuse.test import amusetest
from amuse.units import units, nbody_system
from amuse.ext.molecular_cloud import molecular_cloud, ism_cube
from amuse.ext.evrard_test import sobol_unit_cube

numpy.random.seed(1234567)


class MolecularCloudTests(amusetest.TestCase):
    def test1(self):
        mc = molecular_cloud(targetN=1000, base_grid=sobol_unit_cube).result
        self.assertEqual(len(mc), 995)
        ek = mc.kinetic_energy()
        ep = mc.potential_energy(G=nbody_system.G)
        eth = mc.thermal_energy()
        self.assertAlmostRelativeEqual(eth/ep, -0.01, 2)
        self.assertAlmostRelativeEqual(ek/ep, -1., 2)

    def test2(self):
        mc = molecular_cloud(targetN=1000, ethep_ratio=0.05, base_grid=sobol_unit_cube).result
        self.assertEqual(len(mc), 995)
        ek = mc.kinetic_energy()
        ep = mc.potential_energy(G=nbody_system.G)
        eth = mc.thermal_energy()
        self.assertAlmostRelativeEqual(eth/ep, -0.05, 2)
        self.assertAlmostRelativeEqual(ek/ep, -1., 2)

    def test3(self):
        mc = molecular_cloud(targetN=1000, ekep_ratio=2., base_grid=sobol_unit_cube).result
        self.assertEqual(len(mc), 995)
        ek = mc.kinetic_energy()
        ep = mc.potential_energy(G=nbody_system.G)
        eth = mc.thermal_energy()
        self.assertAlmostRelativeEqual(eth/ep, -0.01, 2)
        self.assertAlmostRelativeEqual(ek/ep, -2., 2)



from amuse.test import amusetest

import os
import numpy
import time

from amuse import io
from amuse.io import store
from amuse.io import store_v1
from amuse.io import store_v2
from amuse.units import units
from amuse.units import nbody_system
from amuse.datamodel import Particles
from amuse.datamodel import ParticlesOverlay
from amuse.datamodel import Grid
from amuse.datamodel import ParticlesSuperset


class _AbstractTestStoreHDF(object):

    def store_factory(self):
        raise NotImplementedError

    def store_version(self):
        raise NotImplementedError

    def get_version_in_store(self, container):
        raise NotImplementedError

    def test1(self):
        test_results_path = self.get_path_to_results()
        output_file = os.path.join(test_results_path, "test1."+self.store_version()+".hdf5")
        if os.path.exists(output_file):
            os.remove(output_file)
        instance = self.store_factory()(output_file)

        number_of_particles = 10
        p = Particles(number_of_particles)
        p.mass = [x * 2.0 for x in range(number_of_particles)] | units.kg
        p.model_time = 2.0 | units.s

        instance.store(p)

        loaded_particles = instance.load()

        loaded_mass_in_kg = loaded_particles.mass.value_in(units.kg)
        previous_mass_in_kg = p.mass.value_in(units.kg)
        for expected, actual in zip(previous_mass_in_kg, loaded_mass_in_kg):
            self.assertEqual(expected, actual)

        instance.close()

    def test2(self):
        test_results_path = self.get_path_to_results()
        output_file = os.path.join(test_results_path, "test2."+self.store_version()+".hdf5")
        if os.path.exists(output_file):
            os.remove(output_file)

        instance = self.store_factory()(output_file)
        number_of_particles = 10
        p = Particles(number_of_particles)
        p.mass = [x * 2.0 for x in range(number_of_particles)] | units.kg
        instance.store(p.savepoint(1 | units.Myr))

        p.mass = [x * 4.0 for x in range(number_of_particles)] | units.kg
        instance.store(p.savepoint(2 | units.Myr))
        instance.close()

        instance = self.store_factory()(output_file)
        loaded_particles = instance.load()
        self.assertAlmostRelativeEquals(self.get_version_in_store(loaded_particles).get_timestamp(), 2 | units.Myr)
        self.assertAlmostRelativeEquals(self.get_version_in_store(loaded_particles).previous_state().get_timestamp(), 1 | units.Myr)
        masses = loaded_particles[1].get_timeline_of_attribute("mass")
        self.assertEqual(len(masses), 2)

        instance.close()

    def test3(self):
        test_results_path = self.get_path_to_results()
        output_file = os.path.join(test_results_path, "test3"+self.store_version()+".hdf5")
        if os.path.exists(output_file):
            os.remove(output_file)

        instance = self.store_factory()(output_file)
        number_of_particles = 10
        p = Particles(number_of_particles)
        p.mass = [x * 2.0 for x in range(number_of_particles)] | nbody_system.mass

        instance.store(p.savepoint(1 | nbody_system.time))
        instance.close()

        instance = self.store_factory()(output_file)
        loaded_particles = instance.load()
        self.assertAlmostRelativeEquals(p.mass[1], 2.0 | nbody_system.mass)

        instance.close()

    def test4(self):
        test_results_path = self.get_path_to_results()
        output_file = os.path.join(test_results_path, "test4"+self.store_version()+".hdf5")
        if os.path.exists(output_file):
            os.remove(output_file)

        particles = Particles(10)
        particles.mass = 1.0 | units.kg
        particles.x = 2.0 | units.m
        x = particles.savepoint(2.0 | units.s)

        io.write_set_to_file(x, output_file, format='amuse', version=self.store_version())

        particles_from_file = io.read_set_from_file(output_file, format='amuse', version=self.store_version(), close_file=True)
        particles_in_memory = particles_from_file.copy()

        self.assertAlmostRelativeEquals(particles_in_memory.mass[1], 1.0 | units.kg)

        particles_in_memory.savepoint(4.0 | units.s)

        self.assertAlmostRelativeEquals(particles_from_file.mass[2], 1.0 | units.kg)

    def test5(self):
        test_results_path = self.get_path_to_results()
        output_file = os.path.join(test_results_path, "test5_grid"+self.store_version()+".hdf5")
        if os.path.exists(output_file):
            os.remove(output_file)
        instance = self.store_factory()(output_file)

        shape = 10, 10, 10
        p = Grid(*shape)
        p.mass = ([x * 2.0 for x in range(p.size)] | units.kg).reshape(shape)
        p.model_time = 2.0 | units.s

        instance.store_grid(p)

        loaded_grid = instance.load_grid()

        self.assertEqual(loaded_grid.shape, shape)

        loaded_mass_in_kg = loaded_grid.mass.value_in(units.kg)
        previous_mass_in_kg = p.mass.value_in(units.kg)
        for expected, actual in zip(previous_mass_in_kg, loaded_mass_in_kg):
            self.assertEqual(expected, actual)

        instance.close()

    def test6(self):
        test_results_path = self.get_path_to_results()
        output_file = os.path.join(test_results_path, "test6_grid"+self.store_version()+".hdf5")
        if os.path.exists(output_file):
            os.remove(output_file)
        instance = self.store_factory()(output_file)

        shape = 10, 10, 10
        p = Grid(*shape)
        p.mass = ([x * 2.0 for x in range(p.size)] | units.kg).reshape(shape)
        p.model_time = 2.0 | units.s

        instance.store_grid(p.savepoint(1 | units.Myr))

        loaded_grid = self.get_version_in_store(instance.load_grid())

        self.assertAlmostRelativeEquals(loaded_grid.get_timestamp(), 1 | units.Myr)

        self.assertEqual(loaded_grid.shape, shape)
        self.assertEqual(loaded_grid[0][0][0].mass, 0 | units.kg)
        self.assertAlmostRelativeEquals(loaded_grid[..., 1, 1].mass, p[..., 1, 1].mass)

        instance.close()

    def test7(self):
        test_results_path = self.get_path_to_results()
        output_file = os.path.join(test_results_path, "test7"+self.store_version()+".hdf5")
        if os.path.exists(output_file):
            os.remove(output_file)
        instance = self.store_factory()(output_file)

        number_of_particles = 10
        p = Particles(number_of_particles)
        p.mass = [x * 2.0 for x in range(number_of_particles)] | units.kg
        p.model_time = 2.0 | units.s

        instance.store(p)
        instance.close()

        instance = self.store_factory()(output_file)
        loaded_particles = self.get_version_in_store(instance.load())

        loaded_particles.mass = [x * 3.0 for x in range(number_of_particles)] | units.kg
        previous_mass_in_kg = [x * 3.0 for x in range(number_of_particles)]
        instance.close()

        instance = self.store_factory()(output_file)
        loaded_particles = instance.load()
        loaded_mass_in_kg = loaded_particles.mass.value_in(units.kg)
        for expected, actual in zip(previous_mass_in_kg, loaded_mass_in_kg):
            self.assertEqual(expected, actual)

        instance.close()

        instance = self.store_factory()(output_file)
        loaded_particles = self.get_version_in_store(instance.load())
        loaded_particles[2].mass = 44 | units.kg
        instance.close()
        instance = self.store_factory()(output_file)
        loaded_particles = self.get_version_in_store(instance.load())
        self.assertEqual(44 | units.kg, loaded_particles[2].mass)
        instance.close()

    def test8(self):
        test_results_path = self.get_path_to_results()
        output_file = os.path.join(test_results_path, "test8"+self.store_version()+".hdf5")
        if os.path.exists(output_file):
            os.remove(output_file)

        instance = self.store_factory()(output_file)
        number_of_particles = 10
        p = Particles(number_of_particles)
        p.mass = numpy.asarray([x * 2.0 for x in range(number_of_particles)])

        instance.store(p.savepoint(1 | nbody_system.time))
        instance.close()

        instance = self.store_factory()(output_file)
        loaded_particles = instance.load()
        self.assertAlmostRelativeEquals(p.mass[1], 2.0)
        instance.close()

    def test9(self):
        test_results_path = self.get_path_to_results()
        output_file = os.path.join(test_results_path, "test9"+self.store_version()+".hdf5")
        if os.path.exists(output_file):
            os.remove(output_file)

        number_of_particles = 10
        p = Particles(number_of_particles)
        p.mass = [x * 2.0 for x in range(number_of_particles)] | units.kg
        p.model_time = 2.0 | units.s

        io.write_set_to_file(
            p,
            output_file,
            "hdf5",
            timestamp=2 | units.Myr,
            scale=1 | units.kg,
            version=self.store_version(),
            close_file=True
        )

        loaded_particles = io.read_set_from_file(
            output_file,
            "hdf5",
            version=self.store_version()
        )
        loaded_particles = self.get_version_in_store(loaded_particles)

        a = loaded_particles.collection_attributes
        self.assertAlmostRelativeEquals(a.timestamp, 2 | units.Myr)
        self.assertAlmostRelativeEquals(a.scale, 1 | units.kg)

    def test10(self):
        test_results_path = self.get_path_to_results()
        output_file = os.path.join(test_results_path, "test10"+self.store_version()+".hdf5")
        if os.path.exists(output_file):
            os.remove(output_file)

        shape = 10, 10, 10
        p = Grid(*shape)
        p.mass = ([x * 2.0 for x in range(p.size)] | units.kg).reshape(shape)
        p.model_time = 2.0 | units.s

        io.write_set_to_file(p, output_file, "hdf5",  timestamp=2 | units.Myr, scale=1 | units.kg, version=self.store_version())

        loaded_particles = io.read_set_from_file(output_file, "hdf5", version=self.store_version())
        loaded_particles = self.get_version_in_store(loaded_particles)
        a = loaded_particles.collection_attributes
        self.assertAlmostRelativeEquals(a.timestamp, 2 | units.Myr)
        self.assertAlmostRelativeEquals(a.scale, 1 | units.kg)

    def test11(self):
        test_results_path = self.get_path_to_results()
        output_file = os.path.join(test_results_path, "test11"+self.store_version()+".hdf5")
        if os.path.exists(output_file):
            os.remove(output_file)

        number_of_particles = 10
        p = Particles(number_of_particles)
        p.mass = [x * 2.0 for x in range(number_of_particles)] | units.kg
        p.model_time = 2.0 | units.s

        io.write_set_to_file(p.savepoint(timestamp=2 | units.Myr, scale=1 | units.kg), output_file, "hdf5", version=self.store_version())

        loaded_particles = io.read_set_from_file(output_file, "hdf5", version=self.store_version())
        loaded_particles = self.get_version_in_store(loaded_particles)
        a = loaded_particles.collection_attributes
        self.assertAlmostRelativeEquals(a.timestamp, 2 | units.Myr)
        self.assertAlmostRelativeEquals(a.scale, 1 | units.kg)

    def test12(self):
        test_results_path = self.get_path_to_results()
        output_file = os.path.join(test_results_path, "test12"+self.store_version()+".hdf5")
        if os.path.exists(output_file):
            os.remove(output_file)

        shape = 10, 10, 10
        p = Grid(*shape)
        p.mass = ([x * 2.0 for x in range(p.size)] | units.kg).reshape(shape)
        p.model_time = 2.0 | units.s

        io.write_set_to_file(p.savepoint(timestamp=2 | units.Myr, scale=1 | units.kg), output_file, "hdf5", version=self.store_version())

        loaded_particles = io.read_set_from_file(output_file, "hdf5", version=self.store_version())
        loaded_particles = self.get_version_in_store(loaded_particles)
        a = loaded_particles.collection_attributes
        self.assertAlmostRelativeEquals(a.timestamp, 2 | units.Myr)
        self.assertAlmostRelativeEquals(a.scale, 1 | units.kg)

    def test13(self):
        test_results_path = self.get_path_to_results()
        output_file = os.path.join(test_results_path, "test13"+self.store_version()+".hdf5")
        if os.path.exists(output_file):
            os.remove(output_file)

        stars = Particles(2)
        stars.x = 1.0 | units.km
        stars.md = [[1, 2, 3], [4, 5, 6]] | units.km

        io.write_set_to_file(stars, output_file, "hdf5", version=self.store_version())

        loaded = io.read_set_from_file(output_file, "hdf5", version=self.store_version())
        self.assertEqual(loaded[0].md, [1, 2, 3] | units.km)
        self.assertEqual(loaded[1].md, [4, 5, 6] | units.km)

        self.assertEqual(loaded.md[0], [1, 2, 3] | units.km)
        self.assertEqual(loaded.md[1], [4, 5, 6] | units.km)

        self.assertEqual(self.get_version_in_store(loaded)[0].md, [1, 2, 3] | units.km)
        # self.assertEquals(self.get_version_in_store(loaded)[0].md,  [7,8,9] | units.km)

    def test14(self):
        test_results_path = self.get_path_to_results()
        output_file = os.path.join(test_results_path, "test14"+self.store_version()+".hdf5")
        if os.path.exists(output_file):
            os.remove(output_file)

        stars = Particles(2)
        stars.x = 1.0 | units.km
        stars.md = [[[1, 3], [2, 4], [3, 5]], [[4, 6], [5, 7], [6, 8]]]

        io.write_set_to_file(stars, output_file, "hdf5", version=self.store_version())

        loaded = io.read_set_from_file(output_file, "hdf5", version=self.store_version())
        self.assertEqual(loaded[0].md, [[1, 3], [2, 4], [3, 5]])
        self.assertEqual(loaded[1].md, [[4, 6], [5, 7], [6, 8]])

        self.assertEqual(loaded.md[0], [[1, 3], [2, 4], [3, 5]])
        self.assertEqual(loaded.md[1], [[4, 6], [5, 7], [6, 8]])

    def test15(self):

        test_results_path = self.get_path_to_results()
        output_file = os.path.join(test_results_path, "test15"+self.store_version()+".hdf5")
        if os.path.exists(output_file):
            os.remove(output_file)

        stars = Particles(2)
        stars.x = 1.0 | units.km
        stars.md = [[[1, 3], [2, 4], [3, 5]], [[4, 6], [5, 7], [6, 8]]]

        io.write_set_to_file(stars, output_file, "hdf5", version=self.store_version())
        processor = self.store_factory()(output_file, True, open_for_writing=True)
        loaded = processor.load()
        self.get_version_in_store(loaded)[0].md = [[3, 1], [3, 4], [5, 2]]
        self.assertEqual(self.get_version_in_store(loaded)[0].md, [[3, 1], [3, 4], [5, 2]])
        processor.close()

        loaded = io.read_set_from_file(output_file, "hdf5", version=self.store_version())
        self.assertEqual(loaded[0].md, [[3, 1], [3, 4], [5, 2]])
        self.assertEqual(loaded[1].md, [[4, 6], [5, 7], [6, 8]])

    def test16(self):
        import h5py
        test_results_path = self.get_path_to_results()
        output_file = os.path.join(test_results_path, "test16"+self.store_version()+".hdf5")
        if os.path.exists(output_file):
            os.remove(output_file)

        stars = Particles(2)
        stars[0].x = 1.0 | units.km
        stars[1].x = 2.0 | units.km
        stars[0].nn = stars[1]
        stars[1].nn = stars[0]

        self.assertEqual(stars[0].nn, stars[1])
        self.assertEqual(stars[1].nn, stars[0])

        io.write_set_to_file(stars, output_file, "hdf5", version=self.store_version())
        processor = self.store_factory()(output_file, True, open_for_writing=True)
        loaded = processor.load()

        self.assertEqual(loaded[0].nn, loaded[1])
        self.assertEqual(loaded[1].nn, loaded[0])
        self.assertEqual(loaded[0].nn.key, stars[1].key)
        self.assertEqual(loaded[0].nn.x, stars[1].x)

    def test17(self):

        test_results_path = self.get_path_to_results()
        output_file = os.path.join(test_results_path, "test17"+self.store_version()+".hdf5")
        if os.path.exists(output_file):
            os.remove(output_file)

        stars = Particles(4)
        stars[0].x = 1.0 | units.km
        stars[1].x = 2.0 | units.km
        stars[2].x = 3.0 | units.km
        stars[3].x = 4.0 | units.km

        binaries = Particles(2)
        binaries[0].y = 1.0 | units.km
        binaries[1].y = 2.0 | units.km
        binaries[0].child1 = stars[0]
        binaries[0].child2 = stars[1]
        binaries[1].child1 = stars[2]
        binaries[1].child2 = stars[3]

        self.assertEqual(binaries[0].child1, stars[0])
        self.assertEqual(binaries[1].child1, stars[2])

        io.write_set_to_file(binaries, output_file, "hdf5", version=self.store_version())

        loaded = io.read_set_from_file(output_file, "hdf5", version=self.store_version())

        self.assertEqual(loaded[0].child1.key, stars[0].key)
        self.assertEqual(loaded[1].child1.key, stars[2].key)

    def test18(self):

        test_results_path = self.get_path_to_results()
        output_file = os.path.join(test_results_path, "test18"+self.store_version()+".hdf5")
        if os.path.exists(output_file):
            os.remove(output_file)

        stars = Particles(2)
        stars[0].x = 1.0 | units.km
        stars[1].x = 2.0 | units.km

        binaries = Particles(1)
        binaries[0].y = 1.0 | units.km
        binaries[0].child1 = stars[0]
        binaries[0].child2 = stars[1]

        stars[0].parent = binaries[0]
        stars[1].parent = binaries[0]

        self.assertEqual(binaries[0].child1, stars[0])
        self.assertEqual(binaries[0].child2, stars[1])
        self.assertEqual(binaries[0].child1.parent, binaries[0])
        self.assertEqual(binaries[0].child2.parent, binaries[0])

        io.write_set_to_file(binaries, output_file, "hdf5", version=self.store_version())

        loaded = io.read_set_from_file(output_file, "hdf5", version=self.store_version())

        self.assertEqual(loaded[0].child1.key, stars[0].key)
        self.assertEqual(loaded[0].child2.key, stars[1].key)
        self.assertEqual(loaded[0].child1, stars[0])
        self.assertEqual(loaded[0].child2, stars[1])
        self.assertEqual(loaded[0].child1.parent, loaded[0])
        self.assertEqual(loaded[0].child2.parent, loaded[0])

    def test19(self):
        test_results_path = self.get_path_to_results()
        output_file = os.path.join(test_results_path, "test19"+self.store_version()+".hdf5")
        if os.path.exists(output_file):
            os.remove(output_file)

        stars = Particles(2)
        stars[0].x = 1.0 | units.km
        stars[1].x = 2.0 | units.km

        binaries = Particles(1)
        binaries[0].y = 1.0 | units.km
        binaries[0].child1 = stars[0]
        binaries[0].child2 = stars[1]

        stars[0].parent = binaries[0]
        stars[1].parent = binaries[0]

        self.assertEqual(binaries[0].child1, stars[0])
        self.assertEqual(binaries[0].child2, stars[1])
        self.assertEqual(binaries[0].child1.parent, binaries[0])
        self.assertEqual(binaries[0].child2.parent, binaries[0])

        io.write_set_to_file([binaries, stars], output_file, "hdf5", names=['binaries', 'children'], version=self.store_version())

        loader_binaries, loaded_stars = io.read_set_from_file(output_file, "hdf5", names=['binaries', 'children'], version=self.store_version())

        self.assertEqual(loader_binaries[0].child1.key, stars[0].key)
        self.assertEqual(loader_binaries[0].child2.key, stars[1].key)
        self.assertEqual(loader_binaries[0].child1, loaded_stars[0])
        self.assertEqual(loader_binaries[0].child2, loaded_stars[1])
        self.assertEqual(loader_binaries[0].child1.parent, loader_binaries[0])
        self.assertEqual(loader_binaries[0].child2.parent, loader_binaries[0])

    def test20(self):

        test_results_path = self.get_path_to_results()
        output_file = os.path.join(test_results_path, "test20"+self.store_version()+".hdf5")
        if os.path.exists(output_file):
            os.remove(output_file)

        shape = 10, 10, 10
        p = Grid(*shape)
        p.mass = (numpy.asarray([x * 2.0 for x in range(p.size)])).reshape(shape)

        io.write_set_to_file(p, output_file, "hdf5", version=self.store_version())

        loaded = io.read_set_from_file(output_file, "hdf5", version=self.store_version())
        loaded = self.get_version_in_store(loaded)

        self.assertAlmostRelativeEquals(p.mass[0][1][2], 24)
        self.assertAlmostRelativeEquals(p[0][1][2].mass, 24)

    def test21(self):

        test_results_path = self.get_path_to_results()
        output_file = os.path.join(test_results_path, "test21"+self.store_version()+".hdf5")
        if os.path.exists(output_file):
            os.remove(output_file)

        stars = Particles(2)
        stars.x = 1.0 | units.km
        stars.md = [[[1, 3], [2, 4], [3, 5]], [[4, 6], [5, 7], [6, 8]]]

        io.write_set_to_file(stars, output_file, "hdf5", version=self.store_version())

        loaded = io.read_set_from_file(output_file, "hdf5", close_file=True, version=self.store_version(), copy_history=False)

        self.assertEqual(loaded[0].md, [[1, 3], [2, 4], [3, 5]])
        self.assertEqual(loaded[1].md, [[4, 6], [5, 7], [6, 8]])

        previous = loaded.previous_state()
        self.assertEqual(previous, None)

    def test22(self):

        test_results_path = self.get_path_to_results()
        output_file = os.path.join(test_results_path, "test22"+self.store_version()+".hdf5")
        if os.path.exists(output_file):
            os.remove(output_file)

        stars = Particles(2)
        stars.x = 1.0 | units.km

        overlay = ParticlesOverlay(stars)
        overlay.y = 2.0 | units.km

        io.write_set_to_file(overlay, output_file, "hdf5", version=self.store_version())

        loaded = io.read_set_from_file(output_file, "hdf5", close_file=True, version=self.store_version())

        self.assertEqual(loaded[0].x,  1.0 | units.km)
        self.assertEqual(loaded[1].y,  2.0 | units.km)

    def test23(self):
        test_results_path = self.get_path_to_results()
        output_file = os.path.join(test_results_path, "test23"+self.store_version()+".hdf5")
        if os.path.exists(output_file):
            os.remove(output_file)

        particles = Particles()

        io.write_set_to_file(particles, output_file, format='amuse', version=self.store_version())

        particles_from_file = io.read_set_from_file(output_file, format='amuse', version=self.store_version())

        self.assertEqual(len(particles_from_file), 0)

    def test24(self):
        test_results_path = self.get_path_to_results()
        output_file = os.path.join(test_results_path, "test24"+self.store_version()+".hdf5")
        if os.path.exists(output_file):
            os.remove(output_file)

        particles = Particles(20)
# ~        particles = Particles(1000000) # For testing memory usage
        particles.mass = 1.0 | units.kg
        particles.position = [0.0, 0.0, 0.0] | units.m

        for i in range(10):
            particles.position += [1.0, 2.0, 3.0] | units.m
            io.write_set_to_file(particles, output_file, format='amuse', version=self.store_version(), append_to_file=True)

        particles_from_file = io.read_set_from_file(output_file, format='amuse', version=self.store_version())
        os.remove(output_file)
        self.assertEqual(len(list(particles_from_file.history)), 10)
        for i, snap in enumerate(particles_from_file.history):
            self.assertEqual(len(snap), 20)
            self.assertEqual((i+1) * ([1.0, 2.0, 3.0] | units.m), snap.center_of_mass())

    def test25(self):
        test_results_path = self.get_path_to_results()
        output_file = os.path.join(test_results_path, "test25"+self.store_version()+".hdf5")
        if os.path.exists(output_file):
            os.remove(output_file)

        particles = Particles(20)
# ~        particles = Particles(1000000) # For testing memory usage
        particles.mass = 1.0 | units.kg
        particles.position = [0.0, 0.0, 0.0] | units.m

        for i in range(10):
            particles.position += [1.0, 2.0, 3.0] | units.m
            io.write_set_to_file(particles, output_file, format='amuse', version=self.store_version(), append_to_file=True)

        particles_from_file = io.read_set_from_file(output_file, format='amuse', version=self.store_version(), copy_history=True, close_file=True)

        history = list(particles_from_file.history)
        self.assertEqual(len(history), 10)
        self.assertFalse("HDF" in str(type(history[1]._private.attribute_storage)))
        for i, snap in enumerate(particles_from_file.history):
            self.assertEqual(len(snap), 20)
            self.assertEqual((i+1) * ([1.0, 2.0, 3.0] | units.m), snap.center_of_mass())
        os.remove(output_file)

    def test26(self):
        test_results_path = self.get_path_to_results()
        output_file = os.path.join(test_results_path, "test26"+self.store_version()+".hdf5")
        if os.path.exists(output_file):
            os.remove(output_file)

        p = Grid(20)

        io.write_set_to_file(p, output_file, format='amuse', version=self.store_version())

        os.remove(output_file)

    def test27(self):
        test_results_path = self.get_path_to_results()
        output_file = os.path.join(test_results_path, "test27"+self.store_version()+".hdf5")
        if os.path.exists(output_file):
            os.remove(output_file)
        instance = self.store_factory()(output_file)

        shape = 10,
        p = Grid(*shape)
        p.mass = ([x * 2.0 for x in range(p.size)] | units.kg).reshape(shape)
        p.model_time = 2.0 | units.s

        instance.store_grid(p)

        loaded_grid = instance.load_grid()

        self.assertEqual(loaded_grid.shape, shape)

        loaded_mass_in_kg = loaded_grid.mass.value_in(units.kg)
        previous_mass_in_kg = p.mass.value_in(units.kg)
        for expected, actual in zip(previous_mass_in_kg, loaded_mass_in_kg):
            self.assertEqual(expected, actual)

        instance.close()

    def test28(self):

        test_results_path = self.get_path_to_results()
        output_file = os.path.join(test_results_path, "test28"+self.store_version()+".hdf5")
        if os.path.exists(output_file):
            os.remove(output_file)

        stars = Particles(2)
        stars[0].x = 1.0 | units.km
        stars[1].x = 2.0 | units.km

        binaries = Particles(1)
        binaries[0].y = 1.0 | units.km
        binaries[0].child1 = stars[0]
        binaries[0].child2 = stars[1]

        stars[0].parent = binaries[0]
        stars[1].parent = binaries[0]

        self.assertEqual(binaries[0].child1, stars[0])
        self.assertEqual(binaries[0].child2, stars[1])
        self.assertEqual(binaries[0].child1.parent, binaries[0])
        self.assertEqual(binaries[0].child2.parent, binaries[0])

        io.write_set_to_file(binaries, output_file, "hdf5", version=self.store_version())

        with io.read_set_from_file(output_file, "hdf5", version=self.store_version(), return_context=True) as loaded:

            self.assertEqual(loaded[0].child1.key, stars[0].key)
            self.assertEqual(loaded[0].child2.key, stars[1].key)
            self.assertEqual(loaded[0].child1, stars[0])
            self.assertEqual(loaded[0].child2, stars[1])
            self.assertEqual(loaded[0].child1.parent, loaded[0])
            self.assertEqual(loaded[0].child2.parent, loaded[0])

    def test29(self):
        test_results_path = self.get_path_to_results()
        output_file = os.path.join(test_results_path, "test29"+self.store_version()+".hdf5")
        if os.path.exists(output_file):
            os.remove(output_file)

        particles = Grid(10)
        particles.attribute1 = "normal"
        particles.attribute2 = "unicode"

        io.write_set_to_file(particles, output_file, format='amuse', version=self.store_version())

        output = io.read_set_from_file(output_file, format='amuse')
        self.assertEqual(output[0].attribute2, "unicode")

    def test30(self):
        test_results_path = self.get_path_to_results()
        output_file = os.path.join(test_results_path, "test30"+self.store_version()+".hdf5")
        if os.path.exists(output_file):
            os.remove(output_file)

        particles = Particles(10)
        particles.mass = 10 | units.kg

        io.write_set_to_file(particles, output_file, format='amuse', version=self.store_version())

        output = io.read_set_from_file(output_file, format='amuse', allow_writing=True)
        output.new_attribute = 2 * output.mass

        self.assertEqual(output.new_attribute, 2 * output.mass)
        output = next(output.iter_history())
        output.new_attribute = 2 * output.mass

        self.assertEqual(output.new_attribute, 2 * output.mass)

    def test31(self):
        test_results_path = self.get_path_to_results()
        output_file = os.path.join(test_results_path, "test31"+self.store_version()+".h5")
        if os.path.exists(output_file):
            os.remove(output_file)

        p = Particles(10)
        p.mass = numpy.arange(10)
        q = Particles(3)
        q[0].link = p[1]
        q[1].link = p[3]
        q[2].link = p[4]
        io.write_set_to_file(q, output_file, "amuse", version=self.store_version())
        r = io.read_set_from_file(output_file, "amuse")
        self.assertEqual(len(r.link.shape), 1)
        self.assertEqual(r[0].link.mass, 1)
        self.assertEqual(r[1].link.mass, 3)
        self.assertEqual(r[2].link.mass, 4)


class TestStoreHDFV1(amusetest.TestCase, _AbstractTestStoreHDF):

    def store_factory(self):
        return store_v1.StoreHDF

    def store_version(self):
        return '1.0'

    def get_version_in_store(self, container):
        return container.previous_state()


class TestStoreHDFV2(amusetest.TestCase, _AbstractTestStoreHDF):

    def store_factory(self):
        return store_v2.StoreHDF

    def store_version(self):
        return '2.0'

    def get_version_in_store(self, container):
        return container

    def test52(self):
        test_results_path = self.get_path_to_results()
        output_file = os.path.join(test_results_path, "test52"+self.store_version()+".hdf5")
        if os.path.exists(output_file):
            os.remove(output_file)

        stars = Particles(2)
        stars[0].x = 1.0 | units.km
        stars[1].x = 2.0 | units.km

        gas = Particles(1)
        gas[0].y = 1.0 | units.km

        stars[0].gas = gas

        io.write_set_to_file(stars, output_file, "hdf5", version=self.store_version())

        loaded_stars = io.read_set_from_file(output_file, "hdf5", version=self.store_version())

        self.assertEqual(loaded_stars[0].gas[0].key, gas[0].key)
        self.assertAlmostRelativeEquals(loaded_stars[0].gas[0].y, 1.0 | units.km)
        self.assertEqual(loaded_stars[1].gas, None)

    def test53(self):
        test_results_path = self.get_path_to_results()
        output_file = os.path.join(test_results_path, "test53"+self.store_version()+".hdf5")
        if os.path.exists(output_file):
            os.remove(output_file)

        stars = Particles(2)
        stars[0].x = 1.0 | units.km
        stars[1].x = 2.0 | units.km

        gas = Grid(2, 3)
        gas.y = [[1.0, 2.0, 3.0], [4.0, 5.0, 6.0]] | units.km

        stars[0].gas = gas

        io.write_set_to_file(stars, output_file, "hdf5", version=self.store_version())
        loaded_stars = io.read_set_from_file(output_file, "hdf5", version=self.store_version())

        self.assertAlmostRelativeEquals(loaded_stars[0].gas[0][0].y, 1.0 | units.km)
        self.assertAlmostRelativeEquals(loaded_stars[0].gas[0][2].y, 3.0 | units.km)
        self.assertEqual(loaded_stars[1].gas, None)

    def test54(self):
        test_results_path = self.get_path_to_results()
        output_file = os.path.join(test_results_path, "test23"+self.store_version()+".hdf5")
        if os.path.exists(output_file):
            os.remove(output_file)

        stars = Particles(2)
        stars[0].x = 1.0 | units.km
        stars[1].x = 2.0 | units.km

        gas = Grid(2, 3)
        gas.y = [[1.0, 2.0, 3.0], [4.0, 5.0, 6.0]] | units.km

        stars[0].gas = gas[0][0]
        stars[1].gas = gas[1][0]

        self.assertAlmostRelativeEquals(stars[0].gas.y, 1.0 | units.km)
        self.assertAlmostRelativeEquals(stars[1].gas.y, 4.0 | units.km)

        io.write_set_to_file(stars, output_file, "hdf5", version=self.store_version())
        loaded_stars = io.read_set_from_file(output_file, "hdf5", version=self.store_version())

        self.assertAlmostRelativeEquals(loaded_stars[0].gas.y, 1.0 | units.km)
        self.assertAlmostRelativeEquals(loaded_stars[1].gas.y, 4.0 | units.km)

    def test54b(self):
        test_results_path = self.get_path_to_results()
        output_file = os.path.join(test_results_path, "test24"+self.store_version()+".hdf5")
        if os.path.exists(output_file):
            os.remove(output_file)

        stars = Particles(2)
        stars[0].x = 1.0 | units.km
        stars[1].x = 2.0 | units.km

        gas = Grid(2, 3)
        gas.y = [[1.0, 2.0, 3.0], [4.0, 5.0, 6.0]] | units.km
        gas[0][0].particle = stars[0]
        gas[1][2].particle = stars[0]
        gas[1][0].particle = stars[1]

        # stars[0].gas = gas[0][0]
        # stars[1].gas = gas[1][0]

        self.assertAlmostRelativeEquals(gas[0][0].particle.x, 1.0 | units.km)
        self.assertAlmostRelativeEquals(gas[1][2].particle.x, 1.0 | units.km)
        self.assertAlmostRelativeEquals(gas[1][0].particle.x, 2.0 | units.km)

        io.write_set_to_file(gas, output_file, "hdf5", version=self.store_version())
        loaded_gas = io.read_set_from_file(output_file, "hdf5", version=self.store_version())

        self.assertAlmostRelativeEquals(loaded_gas[0][0].particle.x, 1.0 | units.km)
        self.assertAlmostRelativeEquals(loaded_gas[1][2].particle.x, 1.0 | units.km)
        self.assertAlmostRelativeEquals(loaded_gas[1][0].particle.x, 2.0 | units.km)
        self.assertEqual(id(loaded_gas[0][0].particle.get_containing_set()), id(loaded_gas[1][2].particle.get_containing_set()))

        self.assertEqual(loaded_gas.particle.shape, loaded_gas.shape)

        gas_copy = loaded_gas.copy()

        gas_copy[0][0].particle.x = 3 | units.km
        self.assertAlmostRelativeEquals(loaded_gas[0][0].particle.x, 1.0 | units.km)
        self.assertAlmostRelativeEquals(gas_copy[0][0].particle.x, 3.0 | units.km)
        self.assertAlmostRelativeEquals(gas_copy[1][2].particle.x, 3.0 | units.km)

    def test55(self):
        test_results_path = self.get_path_to_results()
        output_file = os.path.join(test_results_path, "test25"+self.store_version()+".hdf5")
        if os.path.exists(output_file):
            os.remove(output_file)

        particles1 = Particles(2)
        particles1[0].x = 1.0 | units.km
        particles1[1].x = 2.0 | units.km
        particles2 = Particles(2)
        particles2[0].x = 3.0 | units.km
        particles2[1].x = 4.0 | units.km

        particles_superset = ParticlesSuperset([particles1, particles2])

        self.assertAlmostRelativeEquals(particles_superset[0].x, 1.0 | units.km)
        self.assertAlmostRelativeEquals(particles_superset[2].x, 3.0 | units.km)

        io.write_set_to_file(particles_superset, output_file, "hdf5", version=self.store_version())
        loaded_particles_superset = io.read_set_from_file(output_file, "hdf5", version=self.store_version())

        self.assertAlmostRelativeEquals(loaded_particles_superset[0].x, 1.0 | units.km)
        self.assertAlmostRelativeEquals(loaded_particles_superset[2].x, 3.0 | units.km)

    def test56(self):
        test_results_path = self.get_path_to_results()
        output_file = os.path.join(test_results_path, "test26"+self.store_version()+".hdf5")
        if os.path.exists(output_file):
            os.remove(output_file)

        number_of_particles = 10
        p = Particles(number_of_particles)
        p.mass = [x * 2.0 for x in range(number_of_particles)] | units.kg
        p.model_time = 2.0 | units.s

        gas = Grid(2, 3)
        gas.y = [[1.0, 2.0, 3.0], [4.0, 5.0, 6.0]] | units.km

        io.write_set_to_file(
            p,
            output_file,
            "hdf5",
            particle=p[1],
            particles=p,
            gridpoint=gas[0][0],
            grid=gas,
            version=self.store_version()
        )

        loaded_particles = io.read_set_from_file(
            output_file,
            "hdf5",
            version=self.store_version(), copy_history=False, close_file=False
        )

        attributes = loaded_particles.collection_attributes
        self.assertAlmostRelativeEquals(attributes.particle.mass, loaded_particles[1].mass)
        self.assertEquals(attributes.particle.key, loaded_particles[1].key)
        self.assertEquals(id(attributes.particles), id(loaded_particles))
        self.assertAlmostRelativeEquals(attributes.gridpoint.y, 1.0 | units.km)
        self.assertAlmostRelativeEquals(attributes.grid[0][0].y, 1.0 | units.km)

    def test57(self):
        test_results_path = self.get_path_to_results()
        output_file = os.path.join(test_results_path, "test26"+self.store_version()+".hdf5")
        if os.path.exists(output_file):
            os.remove(output_file)

        x = Particles(2)
        x.sub = None
        x[0].sub = Particles(2)
        io.write_set_to_file(x, output_file, "amuse", version=self.store_version())
        y = io.read_set_from_file(output_file, "amuse")

        self.assertEqual(len(x), len(y))
        self.assertEqual(len(x[0].sub), len(y[0].sub))
        self.assertTrue(y[1].sub == x[1].sub == None)

        os.remove(output_file)

    def test58(self):
        test_results_path = self.get_path_to_results()
        output_file = os.path.join(test_results_path, "test58"+self.store_version()+".h5")
        if os.path.exists(output_file):
            os.remove(output_file)

        x = Particles(keys=(1, 2))
        x.mass = [1, 2] | units.kg
        y = Particles(keys=(11, 12, 13, 14))
        y.mass = [40, 50, 60, 70] | units.kg

        x.sub = None
        x[0].sub = y[0:2]
        y[0].sub = y[2:]

        io.write_set_to_file(x, output_file, "amuse", version=self.store_version())
        z = io.read_set_from_file(output_file, "amuse")

        self.assertEqual(len(x), len(z))
        self.assertEqual(len(x[0].sub), len(z[0].sub))
        self.assertEqual(x[0].sub[0].sub.key, (13, 14))
        self.assertEqual(z[0].sub[0].sub.key, (13, 14))
        self.assertEqual(id(x[0].sub._original_set()), id(x[0].sub[0].sub._original_set()))
        # no more subsets, could this be fixed, would be very nice to do so
        # self.assertEqual(id(z[0].sub._original_set()), id(z[0].sub[0].sub._original_set()))
        self.assertTrue(z[1].sub == x[1].sub == None)

        os.remove(output_file)

    def test59(self):
        test_results_path = self.get_path_to_results()
        output_file = os.path.join(test_results_path, "test59"+self.store_version()+".h5")
        if os.path.exists(output_file):
            os.remove(output_file)

        g = Grid(10, 10)

        p = Particles(2)

        p[0].sub = g[0:5]

        io.write_set_to_file(p, output_file, "amuse", version=self.store_version())
        z = io.read_set_from_file(output_file, "amuse")

        os.remove(output_file)

    def test60(self):
        test_results_path = self.get_path_to_results()
        output_file = os.path.join(test_results_path, "test60"+self.store_version()+".h5")
        if os.path.exists(output_file):
            os.remove(output_file)

        x = Particles(2)
        x.mass = 1 | units.kg
        y = Particles(2)
        y.mass = 2 | units.kg
        io.write_set_to_file((x, y), output_file, "amuse", version=self.store_version(), names=("x", "y"))
        z = io.read_set_from_file(output_file, "amuse")

        os.remove(output_file)
        self.assertTrue("x" in z)
        self.assertTrue("y" in z)
        self.assertAlmostRelativeEquals(z["x"].mass, 1 | units.kg)
        self.assertAlmostRelativeEquals(z["y"].mass, 2 | units.kg)

    def test61(self):
        test_results_path = self.get_path_to_results()
        output_file = os.path.join(test_results_path, "test61"+self.store_version()+".h5")
        if os.path.exists(output_file):
            os.remove(output_file)

        x = Particles(10)
        x.mass = 1 | units.kg
        y = Particles(20)
        y.id = range(20)
        x[0].y = y[5:7]

        io.write_set_to_file(x, output_file, "amuse", version=self.store_version())
        z = io.read_set_from_file(output_file, "amuse")
        self.assertEqual(x[0].y[0].id, 5)
        self.assertEqual(z[0].y[0].id, 5)

        os.remove(output_file)



from amuse.units import units
from amuse.datamodel import Particles, Particle
from amuse.support.exceptions import AmuseException
from amuse.test.amusetest import TestCase

from amuse.ext.sticky_spheres import StickySpheres
from amuse.couple.collision_handler import CollisionHandler


class CollisionCodeForTesting(object):

    class ParametersForTesting(object):
        def __init__(self):
            self.mass_unit = units.kg

    stellar_evolution_code_required = False
    gravity_code_required = False

    def __init__(self, next_mass=1 | units.kg):
        self.next_mass = next_mass
        self.parameters = self.ParametersForTesting()

    def handle_collision(self, primary, secondary, stellar_evolution_code=None):
        result = Particles(1)
        result.mass = self.next_mass.as_quantity_in(self.parameters.mass_unit)
        self.next_mass += 1 | units.kg

        if not stellar_evolution_code is None:
            se_colliders = (primary + secondary).get_intersecting_subset_in(stellar_evolution_code.particles)
            result.radius = se_colliders.radius.sum()

            def get_internal_structure(set, particle=None):
                return dict(mass=result.mass, radius=result.radius)

            result.add_function_attribute("get_internal_structure", None, get_internal_structure)

        return result


class GravityCodeForTesting(object):

    def __init__(self):
        self.particles = Particles(6)
        self.particles.mass = 1 | units.MSun
        self.particles.radius = 3 | units.RSun
        self.particles.position = [[i, 2*i, 3*i] for i in range(6)] | units.AU
        self.particles.velocity = [[i, i**2, i**3] for i in range(6)] | units.km / units.s


class StellarEvolutionCodeForTesting(object):

    def __init__(self, particles=Particles(6)):
        particles.mass = 1 | units.MSun
        particles.radius = range(1, len(particles)+1) | units.RSun
        self.particles = particles


class StellarEvolutionCodeWithInternalStructureForTesting(object):

    def __init__(self, particles=Particles(6)):
        particles.mass = 1 | units.MSun
        particles.radius = range(1, len(particles)+1) | units.RSun
        particles.type = "native star"
        self.particles = particles

    def new_particle_from_model(self, internal_structure, current_age, key=None):
        tmp_star = Particle(key=key)
        tmp_star.mass = internal_structure["mass"]
        tmp_star.radius = internal_structure["radius"]
        tmp_star.type = "new particle from model"
        return self.particles.add_particle(tmp_star)


class TestCollisionHandler(TestCase):

    def test1(self):
        print("Test CollisionHandler with collision code class (creates new instance for each collision)")
        colliders = Particles(2)
        handler = CollisionHandler(CollisionCodeForTesting)

        result = handler.handle_collision(colliders[0], colliders[1])
        self.assertTrue(isinstance(result, Particles))
        self.assertEqual(result.mass, 1 | units.kg)

        result = handler.handle_collision(colliders[0], colliders[1])
        self.assertEqual(result.mass, 1 | units.kg)

    def test2(self):
        print("Test CollisionHandler with collision code instance (same instance for each collision)")
        colliders = Particles(2)
        handler = CollisionHandler(CollisionCodeForTesting())

        result = handler.handle_collision(colliders[0], colliders[1])
        self.assertTrue(isinstance(result, Particles))
        self.assertEqual(result.mass, 1 | units.kg)

        result = handler.handle_collision(colliders[0], colliders[1])
        self.assertEqual(result.mass, 2 | units.kg)

    def test3(self):
        print("Test CollisionHandler with collision code class, arguments and parameters")
        colliders = Particles(2)
        handler = CollisionHandler(
            CollisionCodeForTesting,
            collision_code_arguments=dict(next_mass=5 | units.kg),
            collision_code_parameters=dict(mass_unit=units.g)
        )

        result = handler.handle_collision(colliders[0], colliders[1])
        self.assertTrue(isinstance(result, Particles))
        self.assertEqual(result.mass, 5 | units.kg)
        self.assertTrue(result.mass.unit is units.g)

        result = handler.handle_collision(colliders[0], colliders[1])
        self.assertEqual(result.mass, 5 | units.kg)
        self.assertTrue(result.mass.unit is units.g)

        handler.collision_code_arguments = dict(next_mass=42 | units.kg)
        handler.collision_code_parameters = dict(mass_unit=units.MSun)
        result = handler.handle_collision(colliders[0], colliders[1])
        self.assertEqual(result.mass, 42 | units.kg)
        self.assertTrue(result.mass.unit is units.MSun)

    def test4(self):
        print("Test handle_collisions")
        colliders = Particles(8)
        handler = CollisionHandler(CollisionCodeForTesting)

        result = handler.handle_collisions(colliders[:4], colliders[4:])
        self.assertTrue(isinstance(result, Particles))
        self.assertEqual(len(result), 4)
        self.assertEqual(result.mass, [1, 1, 1, 1] | units.kg)

    def test5(self):
        print("Test CollisionHandler with gravity code")
        gravity = GravityCodeForTesting()
        self.assertEqual(len(gravity.particles), 6)

        handler = CollisionHandler(CollisionCodeForTesting, gravity_code=gravity)
        merged = handler.handle_collisions(gravity.particles[::2], gravity.particles[1::2])
        self.assertTrue(isinstance(merged, Particles))
        self.assertEqual(len(merged), 3)
        self.assertEqual(merged.mass, [1, 1, 1] | units.kg)

        self.assertEqual(len(gravity.particles), 3)
        self.assertEqual(gravity.particles.mass, [1, 1, 1] | units.kg)
        self.assertEqual(gravity.particles.radius, [3, 3, 3] | units.RSun)
        self.assertAlmostEqual(gravity.particles.position,
            [[0.5, 1.0, 1.5], [2.5, 5.0, 7.5], [4.5, 9.0, 13.5]] | units.AU)
        self.assertAlmostEqual(gravity.particles.velocity,
            [[0.5, 0.5, 0.5], [2.5, 6.5, 17.5], [4.5, 20.5, 94.5]] | units.km / units.s)

    def test6(self):
        print("Test CollisionHandler with stellar evolution code, type I")
        stellar_evolution = StellarEvolutionCodeForTesting()
        self.assertEqual(len(stellar_evolution.particles), 6)

        collision_code = CollisionCodeForTesting()
        collision_code.stellar_evolution_code_required = True

        self.assertRaises(AmuseException, CollisionHandler, collision_code, expected_message="CollisionCodeForTesting requires a stellar evolution code: "
             "CollisionHandler(..., stellar_evolution_code=x)")

        handler = CollisionHandler(collision_code, stellar_evolution_code=stellar_evolution)
        merged = handler.handle_collisions(stellar_evolution.particles[::2], stellar_evolution.particles[1::2])
        self.assertTrue(isinstance(merged, Particles))
        self.assertEqual(len(merged), 3)
        self.assertEqual(merged.mass, [1, 2, 3] | units.kg)

        self.assertEqual(len(stellar_evolution.particles), 3)
        self.assertEqual(stellar_evolution.particles.mass, [1, 2, 3] | units.kg)
        self.assertEqual(stellar_evolution.particles.radius, [3, 7, 11] | units.RSun)

    def test7(self):
        print("Test CollisionHandler with stellar evolution code, type II")
        stellar_evolution = StellarEvolutionCodeWithInternalStructureForTesting()
        self.assertEqual(len(stellar_evolution.particles), 6)
        self.assertEqual(stellar_evolution.particles.type, ["native star"]*6)

        collision_code = CollisionCodeForTesting()
        collision_code.stellar_evolution_code_required = True

        self.assertRaises(AmuseException, CollisionHandler, collision_code, expected_message="CollisionCodeForTesting requires a stellar evolution code: "
             "CollisionHandler(..., stellar_evolution_code=x)")

        handler = CollisionHandler(collision_code, stellar_evolution_code=stellar_evolution)
        merged = handler.handle_collisions(stellar_evolution.particles[::2], stellar_evolution.particles[1::2])
        self.assertTrue(isinstance(merged, Particles))
        self.assertEqual(len(merged), 3)
        self.assertEqual(merged.mass, [1, 2, 3] | units.kg)

        self.assertEqual(len(stellar_evolution.particles), 3)
        self.assertEqual(stellar_evolution.particles.mass, [1, 2, 3] | units.kg)
        self.assertEqual(stellar_evolution.particles.radius, [3, 7, 11] | units.RSun)
        self.assertEqual(stellar_evolution.particles.type, ["new particle from model"]*3)

    def test8(self):
        print("Test CollisionHandler with stellar evolution and gravity code")
        gravity = GravityCodeForTesting()
        self.assertEqual(len(gravity.particles), 6)

        stellar_evolution = StellarEvolutionCodeForTesting(particles=gravity.particles.copy())
        self.assertEqual(len(stellar_evolution.particles), 6)

        collision_code = CollisionCodeForTesting()
        collision_code.stellar_evolution_code_required = True

        self.assertRaises(AmuseException, CollisionHandler, collision_code, expected_message="CollisionCodeForTesting requires a stellar evolution code: "
             "CollisionHandler(..., stellar_evolution_code=x)")

        handler = CollisionHandler(
            collision_code,
            stellar_evolution_code=stellar_evolution,
            gravity_code=gravity
        )
        merged = handler.handle_collisions(stellar_evolution.particles[::2], stellar_evolution.particles[1::2])
        self.assertTrue(isinstance(merged, Particles))
        self.assertEqual(len(merged), 3)
        self.assertEqual(merged.mass, [1, 2, 3] | units.kg)

        self.assertEqual(len(stellar_evolution.particles), 3)
        self.assertEqual(stellar_evolution.particles.mass, [1, 2, 3] | units.kg)
        self.assertEqual(stellar_evolution.particles.radius, [3, 7, 11] | units.RSun)

        self.assertEqual(len(gravity.particles), 3)
        self.assertEqual(gravity.particles.mass, [1, 2, 3] | units.kg)
        self.assertEqual(gravity.particles.radius, [3, 7, 11] | units.RSun)
        self.assertAlmostEqual(gravity.particles.position,
            [[0.5, 1.0, 1.5], [2.5, 5.0, 7.5], [4.5, 9.0, 13.5]] | units.AU)
        self.assertAlmostEqual(gravity.particles.velocity,
            [[0.5, 0.5, 0.5], [2.5, 6.5, 17.5], [4.5, 20.5, 94.5]] | units.km / units.s)

    def test9(self):
        print("Test CollisionHandler with gravity code and StickySpheres collision code")
        gravity = GravityCodeForTesting()
        self.assertEqual(len(gravity.particles), 6)
        gravity.particles.mass = [1, 1, 2, 2, 3, 3,] | units.MSun
        gravity.particles.radius = range(101, 107) | units.RSun

        collision_code = StickySpheres(mass_loss=0.1)

        handler = CollisionHandler(
            collision_code,
            gravity_code=gravity
        )
        merged = handler.handle_collisions(gravity.particles[::2], gravity.particles[1::2])
        self.assertTrue(isinstance(merged, Particles))
        self.assertEqual(len(merged), 3)
        self.assertEqual(merged.mass, 0.9 * ([2, 4, 6] | units.MSun))

        self.assertEqual(len(gravity.particles), 3)
        self.assertEqual(gravity.particles.mass, 0.9 * ([2, 4, 6] | units.MSun))
        self.assertEqual(gravity.particles.radius, [102, 104, 106] | units.RSun)
        self.assertAlmostEqual(gravity.particles.position,
            [[0.5, 1.0, 1.5], [2.5, 5.0, 7.5], [4.5, 9.0, 13.5]] | units.AU)
        self.assertAlmostEqual(gravity.particles.velocity,
            [[0.5, 0.5, 0.5], [2.5, 6.5, 17.5], [4.5, 20.5, 94.5]] | units.km / units.s)



from amuse.test.amusetest import TestWithMPI
import os
import sys
import numpy
import math

from amuse.community.huayno.interface import HuaynoInterface, Huayno

from amuse.units import nbody_system
from amuse.units import units, constants
from amuse import datamodel
from amuse.ic import plummer
try:
    from matplotlib import pyplot
    HAS_MATPLOTLIB = True
except ImportError:
    HAS_MATPLOTLIB = False


def elements(starmass, x, y, z, vx, vy, vz, G=constants.G):
    mu = G*starmass
    r = (x**2+y**2+z**2)**0.5
    v2 = (vx**2+vy**2+vz**2)

    e = v2/2-mu/r

    a = -mu/2/e

    hx = y*vz-z*vy
    hy = z*vx-x*vz
    hz = x*vy-y*vx

    rdotv = x*vx+y*vy+z*vz

    ex = v2*x/mu-rdotv*vx/mu-x/r
    ey = v2*y/mu-rdotv*vy/mu-y/r
    ez = v2*z/mu-rdotv*vz/mu-z/r

    h2 = (hx**2+hy**2+hz**2)

    eps = (1-h2/mu/a)**0.5

    return a, eps


def energy(mass, eps, orbiters):
    return (0.5*orbiters.velocity.lengths()**2-constants.G*mass/(orbiters.position.lengths()**2+eps**2)**0.5)


def angular_momentum(orbiters):
    return orbiters.position.cross(orbiters.velocity).lengths()


class TestHuaynoInterface(TestWithMPI):

    def test1(self):
        instance = HuaynoInterface()
        instance.initialize_code()

        res1 = instance.new_particle(mass=11.0, radius=2.0, x=0.0, y=0.0, z=0.0, vx=0.0, vy=0.0, vz=0.0)
        res2 = instance.new_particle(mass=21.0, radius=5.0, x=10.0, y=0.0, z=0.0, vx=10.0, vy=0.0, vz=0.0)

        self.assertEqual(0, res1['index_of_the_particle'])
        self.assertEqual(1, res2['index_of_the_particle'])

        retrieved_state1 = instance.get_state(0)
        retrieved_state2 = instance.get_state(1)

        self.assertEqual(11.0,  retrieved_state1['mass'])
        self.assertEqual(21.0,  retrieved_state2['mass'])
        self.assertEqual(0.0,  retrieved_state1['x'])
        self.assertEqual(10.0,  retrieved_state2['x'])

        instance.cleanup_code()
        instance.stop()

    def test2(self):
        instance = HuaynoInterface()
        instance.initialize_code()

        for i in [0, 1, 2]:
            temp_particle = instance.new_particle(mass=i, radius=1.0, x=0.0, y=0.0, z=0.0, vx=0.0, vy=0.0, vz=0.0)
            self.assertEqual(i, temp_particle['index_of_the_particle'])

        instance.delete_particle(1)

        self.assertEqual(2, instance.get_number_of_particles()['number_of_particles'])

        self.assertEqual(0, instance.get_index_of_first_particle()['index_of_the_particle'])

        self.assertEqual(2, instance.get_index_of_next_particle(0)['index_of_the_next_particle'])
        self.assertEqual(0, instance.get_index_of_next_particle(0)['__result'])
        self.assertEqual(-2, instance.get_index_of_next_particle(1)['__result'])
        self.assertEqual(1, instance.get_index_of_next_particle(2)['__result'])
        instance.cleanup_code()
        instance.stop()

    def test3(self):
        huayno = HuaynoInterface()
        huayno.set_eps2(0.101)
        self.assertEqual(0.101, huayno.get_eps2()['eps2'])
        huayno.set_eps2(0.110)
        self.assertEqual(0.110, huayno.get_eps2()['eps2'])
        huayno.cleanup_code()
        huayno.stop()

    def test5(self):
        huayno = HuaynoInterface()
        huayno.initialize_code()

        huayno.new_particle([10, 20], [0, 0], [0, 0], [0, 0], [0, 0], [0, 0], [0, 0], [1, 1])
        retrieved_state = huayno.get_state(0)

        self.assertEqual(10.0,  retrieved_state['mass'])
        self.assertEqual(1, retrieved_state['radius'])

        retrieved_state = huayno.get_state([0, 1])
        self.assertEqual(20.0,  retrieved_state['mass'][1])
        self.assertEqual(huayno.get_number_of_particles()['number_of_particles'], 2)
        huayno.cleanup_code()
        huayno.stop()


class TestHuayno(TestWithMPI):
    def new_system_of_sun_and_earth(self):
        particles = datamodel.Particles(2)
        particles.mass = [1.0, 3.0037e-6] | units.MSun
        particles.radius = 1.0 | units.RSun
        particles.position = [[0.0, 0.0, 0.0], [1.0, 0.0, 0.0]] | units.AU
        particles.velocity = [[0.0, 0.0, 0.0], [0.0, 0.0, 0.0]] | units.km / units.s
        particles[1].vy = (constants.G * particles.total_mass() / (1.0 | units.AU)).sqrt()
        particles.move_to_center()
        return particles

    def test1(self):
        convert_nbody = nbody_system.nbody_to_si(1.0 | units.MSun, 149.5e6 | units.km)

        huayno = Huayno(convert_nbody)
        huayno.initialize_code()
        huayno.parameters.epsilon_squared = 0.0 | units.AU**2

        stars = self.new_system_of_sun_and_earth()
        earth = stars[1]

        huayno.particles.add_particles(stars)

        huayno.evolve_model(365.0 | units.day)
        huayno.particles.copy_values_of_all_attributes_to(stars)

        position_at_start = earth.position.value_in(units.AU)[0]
        position_after_full_rotation = earth.position.value_in(units.AU)[0]
        self.assertAlmostEqual(position_at_start, position_after_full_rotation, 6)

        huayno.evolve_model(365.0 + (365.0 / 2) | units.day)

        huayno.particles.copy_values_of_all_attributes_to(stars)
        position_after_half_a_rotation = earth.position.value_in(units.AU)[0]
        self.assertAlmostEqual(-position_at_start, position_after_half_a_rotation, 2)

        huayno.evolve_model(365.0 + (365.0 / 2) + (365.0 / 4) | units.day)

        huayno.particles.copy_values_of_all_attributes_to(stars)
        position_after_half_a_rotation = earth.position.value_in(units.AU)[1]
        self.assertAlmostEqual(-position_at_start, position_after_half_a_rotation, 3)

        huayno.cleanup_code()

        huayno.stop()

    def test2(self):
        convert_nbody = nbody_system.nbody_to_si(1.0 | units.MSun, 149.5e6 | units.km)

        instance = Huayno(convert_nbody)
        instance.initialize_code()
        instance.parameters.epsilon_squared = 0.0 | units.AU**2

        stars = self.new_system_of_sun_and_earth()
        earth = stars[1]
        instance.particles.add_particles(stars)

        for x in range(1, 2000, 10):
            instance.evolve_model(x | units.day)
            instance.particles.copy_values_of_all_attributes_to(stars)
            stars.savepoint()

        if HAS_MATPLOTLIB:
            figure = pyplot.figure()
            plot = figure.add_subplot(1, 1, 1)

            x_points = earth.get_timeline_of_attribute("x")
            y_points = earth.get_timeline_of_attribute("y")

            x_points_in_AU = [t_x[1].value_in(units.AU) for t_x in x_points]
            y_points_in_AU = [t_x1[1].value_in(units.AU) for t_x1 in y_points]

            plot.scatter(x_points_in_AU, y_points_in_AU, color="b", marker='o')

            plot.set_xlim(-1.5, 1.5)
            plot.set_ylim(-1.5, 1.5)

            test_results_path = self.get_path_to_results()
            output_file = os.path.join(test_results_path, "huayno-earth-sun2.svg")
            figure.savefig(output_file)

        instance.cleanup_code()
        del instance

    def test3(self):
        convert_nbody = nbody_system.nbody_to_si(1.0 | units.MSun, 149.5e6 | units.km)

        instance = Huayno(convert_nbody)
        instance.initialize_code()
        instance.parameters.epsilon_squared = 0.00001 | units.AU**2

        stars = datamodel.Stars(2)
        star1 = stars[0]
        star2 = stars[1]

        star1.mass = units.MSun(1.0)
        star1.position = units.AU(numpy.array((-1.0, 0.0, 0.0)))
        star1.velocity = units.AUd(numpy.array((0.0, 0.0, 0.0)))
        star1.radius = units.RSun(1.0)

        star2.mass = units.MSun(1.0)
        star2.position = units.AU(numpy.array((1.0, 0.0, 0.0)))
        star2.velocity = units.AUd(numpy.array((0.0, 0.0, 0.0)))
        star2.radius = units.RSun(100.0)

        instance.particles.add_particles(stars)

        for x in range(1, 2000, 10):
            instance.evolve_model(x | units.day)
            instance.particles.copy_values_of_all_attributes_to(stars)
            stars.savepoint()

    def test4(self):
        convert_nbody = nbody_system.nbody_to_si(5.0 | units.kg, 10.0 | units.m)

        instance = Huayno(convert_nbody)
        instance.initialize_code()

        particles = datamodel.Particles(2)
        self.assertEqual(len(instance.particles), 0)

        particles.mass = [15.0, 30.0] | units.kg
        particles.radius = [10.0, 20.0] | units.m
        particles.position = [[10.0, 20.0, 30.0], [20.0, 40.0, 60.0]] | units.m
        particles.velocity = [[0.0, 0.0, 0.0], [0.0, 0.0, 0.0]] | units.m / units.s

        instance.particles.add_particles(particles)
        self.assertEqual(len(instance.particles), 2)

        instance.particles.mass = [17.0, 33.0] | units.kg

        self.assertEqual(instance.get_mass(0), 17.0 | units.kg)
        self.assertEqual(instance.get_mass(1), 33.0 | units.kg)

    def test5(self):
        convert_nbody = nbody_system.nbody_to_si(5.0 | units.kg, 10.0 | units.m)

        instance = Huayno(convert_nbody)
        instance.initialize_code()

        particles = datamodel.Particles(2)
        self.assertEqual(len(instance.particles), 0)

        particles.mass = [15.0, 30.0] | units.kg
        particles.radius = [10.0, 20.0] | units.m
        particles.position = [[10.0, 20.0, 30.0], [20.0, 40.0, 60.0]] | units.m
        particles.velocity = [[0.0, 0.0, 0.0], [0.0, 0.0, 0.0]] | units.m / units.s

        instance.particles.add_particles(particles)
        self.assertEqual(len(instance.particles), 2)

        instance.set_state(1, 16 | units.kg, 20.0 | units.m, 40.0 | units.m, 60.0 | units.m,
                                 1.0 | units.ms, 1.0 | units.ms, 1.0 | units.ms)

        curr_state = instance.get_state(1)
        for expected, actural in zip((16 | units.kg, 20.0 | units.m, 40.0 | units.m, 60.0 | units.m,
                                 1.0 | units.ms, 1.0 | units.ms, 1.0 | units.ms, 0 | units.m), curr_state):
            self.assertAlmostRelativeEquals(actural, expected)

        instance.set_state(1, 16 | units.kg, 20.0 | units.m, 40.0 | units.m, 60.0 | units.m,
                                 1.0 | units.ms, 1.0 | units.ms, 1.0 | units.ms, 20.0 | units.m)

        curr_state = instance.get_state(1)
        for expected, actural in zip((16 | units.kg, 20.0 | units.m, 40.0 | units.m, 60.0 | units.m,
                                 1.0 | units.ms, 1.0 | units.ms, 1.0 | units.ms, 20 | units.m), curr_state):
            self.assertAlmostRelativeEquals(actural, expected)

    def test6(self):
        print("Test6: Testing Huayno parameters")
        convert_nbody = nbody_system.nbody_to_si(1.0 | units.yr, 1.0 | units.AU)
        instance = Huayno(convert_nbody)

        (value, error) = instance.legacy_interface.get_eps2()
        self.assertEqual(0, error)
        self.assertEqual(0.0, value)
        self.assertAlmostEqual(0.0 | units.AU**2, instance.parameters.epsilon_squared, in_units=units.AU**2)
        for x in [0.01, 0.1, 0.2]:
            instance.parameters.epsilon_squared = x | units.AU**2
            self.assertAlmostEqual(x | units.AU**2, instance.parameters.epsilon_squared, in_units=units.AU**2)

        (value, error) = instance.legacy_interface.get_time()
        self.assertEqual(0, error)
        self.assertEqual(0.0, value)
        self.assertAlmostEqual(0.0 | units.yr, instance.parameters.begin_time, in_units=units.yr)
        for x in [1.0, 10.0, 100.0]:
            instance.parameters.begin_time = x | units.yr
            self.assertAlmostEqual(x | units.yr, instance.parameters.begin_time, in_units=units.yr)
        instance.stop()

    def test7(self):
        print("Test7: Testing effect of Huayno parameter epsilon_squared")
        convert_nbody = nbody_system.nbody_to_si(1.0 | units.MSun, 1.0 | units.AU)

        particles = datamodel.Particles(2)
        sun = particles[0]
        sun.mass = 1.0 | units.MSun
        sun.position = [0.0, 0.0, 0.0] | units.AU
        sun.velocity = [0.0, 0.0, 0.0] | units.AU / units.yr
        sun.radius = 1.0 | units.RSun

        earth = particles[1]
        earth.mass = 5.9736e24 | units.kg
        earth.radius = 6371.0 | units.km
        earth.position = [0.0, 1.0, 0.0] | units.AU
        earth.velocity = [2.0*numpy.pi, -0.0001, 0.0] | units.AU / units.yr

        initial_direction = math.atan((earth.velocity[0]/earth.velocity[1]))
        final_direction = []
        for log_eps2 in range(-9, 10, 2):
            instance = Huayno(convert_nbody)
            instance.initialize_code()
            instance.parameters.epsilon_squared = 10.0**log_eps2 | units.AU ** 2
            instance.particles.add_particles(particles)
            instance.commit_particles()
            instance.evolve_model(0.25 | units.yr)
            final_direction.append(math.atan((instance.particles[1].velocity[0] /
                instance.particles[1].velocity[1])))
            instance.stop()
        # Small values of epsilon_squared should result in normal earth-sun dynamics: rotation of 90 degrees
        self.assertAlmostEqual(abs(final_direction[0]), abs(initial_direction+math.pi/2.0), 2)
        # Large values of epsilon_squared should result in ~ no interaction
        self.assertAlmostEqual(final_direction[-1], initial_direction, 2)
        # Outcome is most sensitive to epsilon_squared when epsilon_squared = d(earth, sun)^2
        delta = [abs(final_direction[i+1]-final_direction[i]) for i in range(len(final_direction)-1)]
        self.assertEqual(delta[len(final_direction)//2 - 1], max(delta))

    def test13(self):
        particles = plummer.new_plummer_model(31)

        instance = Huayno(number_of_workers=1)  # , debugger="xterm")
        instance.initialize_code()
        instance.parameters.epsilon_squared = 0.01 | nbody_system.length ** 2
        instance.particles.add_particles(particles)

        instance.evolve_model(0.1 | nbody_system.time)
        instance.synchronize_model()
        expected_positions = instance.particles.position
        instance.stop()

    def test14(self):
        import hashlib

        numpy.random.seed(123456)
        particles = plummer.new_plummer_model(32)
        sha = hashlib.sha1()

        test_set = ["CONSTANT", "SHARED2", "EXTRAPOLATE",
            "PASS_KDK", "PASS_DKD", "HOLD_KDK", "HOLD_DKD",
            "PPASS_DKD", "BRIDGE_KDK", "BRIDGE_DKD",
            "CC", "CC_KEPLER", "CC_BS", "CC_BSA",
            "OK", "SHAREDBS", "SHARED4", "SHARED6", "SHARED8",
            "SHARED10"]

        # ~ test_set=["CONSTANT"]

        for itype in test_set:
            # ~ print()
            # ~ print(itype)
            instance = Huayno()
            instance.parameters.inttype_parameter = Huayno.all_inttypes[itype]
            # ~ instance.parameters.accelerate_zero_mass=False
            instance.particles.add_particles(particles)
            E1 = instance.kinetic_energy+instance.potential_energy
            instance.evolve_model(0.125 | nbody_system.time)
            E2 = instance.kinetic_energy+instance.potential_energy
            if itype != "CONSTANT":
                self.assertLess(abs(E2-E1).number, 1.e-5)
            # ~ print((E2-E1).number)

            part_out = instance.particles.copy()
            position = part_out.position.number
            if hasattr(position, 'tobytes'):
                as_bytes = position.tobytes()
            else:
                as_bytes = numpy.array(position.data, copy=True, order='C')
            sha.update(as_bytes)

            instance.stop()

        # this result is probably dependent on system architecture hence no good for assert
        print()
        print(sha.hexdigest())
        print("4f2aac4d8761f3b07545dcea53f1a65f84a5275b")

    def test14b(self):
        import hashlib

        numpy.random.seed(123456)
        particles = plummer.new_plummer_model(32)
        p2 = plummer.new_plummer_model(32)
        p2.mass *= 0
        sha = hashlib.sha1()

        test_set = ["CONSTANT", "SHARED2", "EXTRAPOLATE",
            "PASS_KDK", "PASS_DKD", "HOLD_KDK", "HOLD_DKD",
            "PPASS_DKD", "BRIDGE_KDK", "BRIDGE_DKD",
            "CC", "CC_KEPLER", "CC_BS", "CC_BSA",
            "OK", "SHAREDBS", "SHARED4", "SHARED6", "SHARED8",
            "SHARED10"]

        for itype in test_set:
            # ~ print()
            # ~ print(itype)
            instance = Huayno(redirection="none")
            instance.parameters.inttype_parameter = Huayno.all_inttypes[itype]
            instance.parameters.accelerate_zero_mass = True
            instance.particles.add_particles(particles)
            instance.particles.add_particles(p2)
            E1 = instance.kinetic_energy+instance.potential_energy
            instance.evolve_model(0.125 | nbody_system.time)
            E2 = instance.kinetic_energy+instance.potential_energy
            if itype != "CONSTANT":
                self.assertLess((E2-E1).number, 1.e-5)

            part_out = instance.particles.copy()
            position = part_out.position.number
            if hasattr(position, 'tobytes'):
                as_bytes = position.tobytes()
            else:
                as_bytes = numpy.array(position.data, copy=True, order='C')
            sha.update(as_bytes)

            instance.stop()

        # this result is probably dependent on system architecture hence no good for assert
        print()
        print(sha.hexdigest())
        print("f3563453fb9b959d8337c6e57a526e1ff52572a7")

    def test14c(self):
        import hashlib

        numpy.random.seed(123456)
        particles = plummer.new_plummer_model(16)
        particles.mass *= 0.25
        p2 = plummer.new_plummer_model(16)
        p2.mass *= 0.75
        sha = hashlib.sha1()

        test_set = ["CONSTANT", "SHARED2", "EXTRAPOLATE",
            "PASS_KDK", "PASS_DKD", "HOLD_KDK", "HOLD_DKD",
            "PPASS_DKD", "BRIDGE_KDK", "BRIDGE_DKD",
            "CC", "CC_KEPLER", "CC_BS", "CC_BSA",
            "OK", "SHAREDBS", "SHARED4", "SHARED6", "SHARED8",
            "SHARED10"]

        # ~ test_set=["CC_BS"]

        for itype in test_set:
            # ~ print()
            # ~ print(itype)
            instance = Huayno(redirection="none")
            instance.parameters.inttype_parameter = Huayno.all_inttypes[itype]
            instance.parameters.accelerate_zero_mass = True
            instance.particles.add_particles(particles)
            instance.particles.add_particles(p2)
            E1 = instance.kinetic_energy+instance.potential_energy
            instance.evolve_model(0.125 | nbody_system.time)
            E2 = instance.kinetic_energy+instance.potential_energy
            # ~ if itype!="CONSTANT":
                # ~ self.assertLess(abs(E2-E1).number, 1.e-5)
            # ~ print((E2-E1).number)
            part_out = instance.particles.copy()
            position = part_out.position.number
            if hasattr(position, 'tobytes'):
                as_bytes = position.tobytes()
            else:
                as_bytes = numpy.array(position.data, copy=True, order='C')
            sha.update(as_bytes)

            instance.stop()

        # this result is probably dependent on system architecture hence no good for assert
        print()
        print(sha.hexdigest())
        print("0af7eaea472989ea4be1bb0e2b8633d6b9af41e4")

    def test15(self):
        particles = plummer.new_plummer_model(512)
        expected_positions = None
        for mode in ["cpu", "openmp", "opencl"]:
            try:
                instance = Huayno(mode=mode, number_of_workers=1)  # , debugger="xterm")
            except:
                print("Running huayno with mode=", mode, " was unsuccessful.")
                continue
            else:
                print("Running huayno with mode=", mode, "... ")

            instance.initialize_code()
            instance.parameters.epsilon_squared = 0.01 | nbody_system.length ** 2
            instance.particles.add_particles(particles)

            instance.evolve_model(0.2 | nbody_system.time)
            instance.synchronize_model()
            if expected_positions is None:
                expected_positions = instance.particles.position
            else:
                self.assertAlmostRelativeEquals(expected_positions, instance.particles.position, 8)
            instance.stop()

    def test16(self):
        instance = Huayno()
        instance.parameters.epsilon_squared = 0.0 | nbody_system.length**2

        particles = datamodel.Particles(2)
        particles.mass = [1.] | nbody_system.mass
        particles.radius = [0.0] | nbody_system.length
        particles.position = [[0.0, 0.0, 0.0], [1.0, 0.0, 0.0]] | nbody_system.length
        particles.velocity = [[0.0, 0.0, 0.0]] | nbody_system.speed
        instance.particles.add_particles(particles)

        zero = 0.0 | nbody_system.length

        for x in (0.25, 0.5, 0.75):
            x0 = x | nbody_system.length
            potential0 = instance.get_potential_at_point(zero, x0, zero, zero)
            fx0, fy0, fz0 = instance.get_gravity_at_point(zero, x0, zero, zero)

            self.assertAlmostEqual(fy0, 0.0 | nbody_system.acceleration, 14)
            self.assertAlmostEqual(fz0, 0.0 | nbody_system.acceleration, 14)

            fx = (-1.0 / (x0**2)+1.0 / (((1.0 | nbody_system.length)-x0)**2)) * (1.0 | nbody_system.length ** 3 / nbody_system.time ** 2)
            self.assertAlmostEqual(fx, fx0, 14)
            self.assertAlmostEqual(potential0, -nbody_system.G*(1. | nbody_system.mass)*(1./x0+1./((1. | nbody_system.length)-x0)), 14)
        instance.stop()

    def _compare_integrator_with_collision_integrator(self, inttype_parameter1, inttype_parameter2):
        numpy.random.seed(12345)
        particles = plummer.new_plummer_model(101)
        instance = Huayno()
        instance.parameters.inttype_parameter = inttype_parameter1
        instance.particles.add_particles(particles)
        instance.evolve_model(0.2 | nbody_system.time)
        expected_position = instance.particles.position
        expected_velocity = instance.particles.velocity
        instance.reset()
        instance.parameters.inttype_parameter = inttype_parameter2
        instance.particles.add_particles(particles)
        instance.evolve_model(0.2 | nbody_system.time)
        self.assertAlmostRelativeEquals(expected_position, instance.particles.position, 8)
        self.assertAlmostRelativeEquals(expected_velocity, instance.particles.velocity, 8)
        instance.stop()

    def _run_collision_with_integrator(self, inttype_parameter):
        particles = datamodel.Particles(7)
        particles.mass = 0.001 | nbody_system.mass
        particles.radius = 0.01 | nbody_system.length
        particles.x = [-101.0, -100.0, -0.5, 0.5, 100.0, 101.0, 104.0] | nbody_system.length
        particles.y = 0 | nbody_system.length
        particles.z = 0 | nbody_system.length
        particles.velocity = [[2, 0, 0], [-2, 0, 0]]*3 + [[-4, 0, 0]] | nbody_system.speed

        instance = Huayno()
        instance.parameters.inttype_parameter = inttype_parameter
        instance.particles.add_particles(particles)
        collisions = instance.stopping_conditions.collision_detection
        collisions.enable()
        instance.evolve_model(1.0 | nbody_system.time)

        self.assertTrue(collisions.is_set())
        self.assertTrue(instance.model_time < 0.5 | nbody_system.time)
        self.assertEqual(len(collisions.particles(0)), 3)
        self.assertEqual(len(collisions.particles(1)), 3)
        self.assertEqual(len(particles - collisions.particles(0) - collisions.particles(1)), 1)
        self.assertEqual(abs(collisions.particles(0).x - collisions.particles(1).x) <=
                (collisions.particles(0).radius + collisions.particles(1).radius),
                [True, True, True])

        sticky_merged = datamodel.Particles(len(collisions.particles(0)))
        sticky_merged.mass = collisions.particles(0).mass + collisions.particles(1).mass
        sticky_merged.radius = collisions.particles(0).radius
        for p1, p2, merged in zip(collisions.particles(0), collisions.particles(1), sticky_merged):
            merged.position = (p1 + p2).center_of_mass()
            merged.velocity = (p1 + p2).center_of_mass_velocity()

        instance.particles.remove_particles(collisions.particles(0) + collisions.particles(1))
        instance.particles.add_particles(sticky_merged)

        instance.evolve_model(1.0 | nbody_system.time)
        self.assertTrue(collisions.is_set())
        self.assertTrue(instance.model_time < 1.0 | nbody_system.time)
        self.assertEqual(len(collisions.particles(0)), 1)
        self.assertEqual(len(collisions.particles(1)), 1)
        self.assertEqual(len(instance.particles - collisions.particles(0) - collisions.particles(1)), 2)
        self.assertEqual(abs(collisions.particles(0).x - collisions.particles(1).x) <=
                (collisions.particles(0).radius + collisions.particles(1).radius),
                [True])
        instance.stop()

    def test17(self):
        print("Compare the SHARED2 integrator with the collision-detection enabled SHARED2_COLLISIONS integrator")
        self._compare_integrator_with_collision_integrator(Huayno.inttypes.SHARED2_COLLISIONS, Huayno.inttypes.SHARED2)
        print("Testing Huayno collision_detection with SHARED2_COLLISIONS")
        self._run_collision_with_integrator(Huayno.inttypes.SHARED2_COLLISIONS)

    def test18(self):
        print("Compare the SHARED4 integrator with the collision-detection enabled SHARED4_COLLISIONS integrator")
        self._compare_integrator_with_collision_integrator(Huayno.inttypes.SHARED4_COLLISIONS, Huayno.inttypes.SHARED4)
        print("Testing Huayno collision_detection with SHARED4_COLLISIONS")
        self._run_collision_with_integrator(Huayno.inttypes.SHARED4_COLLISIONS)

    def test19(self):
        print("Compare the SHARED6 integrator with the collision-detection enabled SHARED6_COLLISIONS integrator")
        self._compare_integrator_with_collision_integrator(Huayno.inttypes.SHARED6_COLLISIONS, Huayno.inttypes.SHARED6)
        print("Testing Huayno collision_detection with SHARED6_COLLISIONS")
        self._run_collision_with_integrator(Huayno.inttypes.SHARED6_COLLISIONS)

    def test20(self):
        print("Compare the SHARED8 integrator with the collision-detection enabled SHARED8_COLLISIONS integrator")
        self._compare_integrator_with_collision_integrator(Huayno.inttypes.SHARED8_COLLISIONS, Huayno.inttypes.SHARED8)
        print("Testing Huayno collision_detection with SHARED8_COLLISIONS")
        self._run_collision_with_integrator(Huayno.inttypes.SHARED8_COLLISIONS)

    def test21(self):
        print("Compare the SHARED10 integrator with the collision-detection enabled SHARED10_COLLISIONS integrator")
        self._compare_integrator_with_collision_integrator(Huayno.inttypes.SHARED10_COLLISIONS, Huayno.inttypes.SHARED10)
        print("Testing Huayno collision_detection with SHARED10_COLLISIONS")
        self._run_collision_with_integrator(Huayno.inttypes.SHARED10_COLLISIONS)

    def test22(self):
        print("Testing zero-mass test particles in Huayno, can be used for removing particles when inside recursive evolve loop")
        sun_and_earth = self.new_system_of_sun_and_earth()
        period = (4.0 * math.pi**2 * (1.0 | units.AU)**3 / (constants.G * sun_and_earth.total_mass())).sqrt()
        convert_nbody = nbody_system.nbody_to_si(1.0 | units.MSun, 1.0 | units.AU)
        huayno = Huayno(convert_nbody)
        huayno.parameters.epsilon_squared = 0.0 | units.AU**2
        huayno.parameters.inttype_parameter = huayno.inttypes.SHARED8

        test_particle = datamodel.Particle(mass=0 | units.MSun, position=[4, 0, 0] | units.AU, velocity=[0, 0, 0] | units.kms)
        test_particle.vy = (constants.G * sun_and_earth.total_mass() / (4.0 | units.AU)).sqrt()
        sun_and_earth.add_particle(test_particle)
        huayno.particles.add_particles(sun_and_earth)
        huayno.evolve_model(period)
        self.assertAlmostRelativeEqual(huayno.particles[:2].x, sun_and_earth[:2].x, 13)
        huayno.evolve_model(1.25 * period)
        self.assertAlmostRelativeEqual(huayno.particles[1].y, sun_and_earth[1].x, 13)
        huayno.evolve_model(8.0 * period)
        self.assertAlmostRelativeEqual(huayno.particles.x, sun_and_earth.x, 8)
        huayno.stop()

    def test23(self):
        print("testing removing and adding particles repeatedly")
        N = 1100
        p1 = plummer.new_plummer_model(N)
        p2 = plummer.new_plummer_model(N)
        h1 = Huayno()
        h1.particles.add_particles(p1[:N//2])
        h1.particles.add_particles(p2[-N//2:])

        h2 = Huayno()
        h2.particles.add_particles(p1)
        h2.particles.remove_particles(p1[N//2:])
        h2.particles.add_particles(p2)
        h2.particles.remove_particles(p2[:-N//2])

        self.assertEqual(len(h1.particles), len(h2.particles))
        self.assertAlmostEqual(h1.kinetic_energy, h2.kinetic_energy, 15)
        self.assertAlmostEqual(h1.potential_energy, h2.potential_energy, 15)

    def test24(self):
        print("test massless particles/ kepler integrator")
        N = 20
        tend = 2. | units.yr
        numpy.random.seed(12345)
        conv = nbody_system.nbody_to_si(4. | units.MSun, 5. | units.AU)
        orbiters = plummer.new_plummer_model(N, conv)
        sun = datamodel.Particle(mass=1. | units.MSun)
        sun.position = [12.3, 1., -.2] | units.AU
        sun.velocity = [10, 50., -20.] | units.kms
        orbiters.mass *= 0.

        a0, eps0 = elements(sun.mass, orbiters.x, orbiters.y, orbiters.z,
                      orbiters.vx, orbiters.vy, orbiters.vz)

        orbiters.position += sun.position
        orbiters.velocity += sun.velocity

        pos = dict()
        for inttype in [20, 14]:
            code = Huayno(conv)
            code.parameters.inttype_parameter = inttype
            code.parameters.timestep_parameter = 0.1
            code.particles.add_particle(sun)
            orbiters2 = code.particles.add_particles(orbiters).copy()
            orbiters2.position -= sun.position
            orbiters2.velocity -= sun.velocity
            code.evolve_model(tend)
            a, eps = elements(sun.mass, orbiters2.x, orbiters2.y, orbiters2.z,
                      orbiters2.vx, orbiters2.vy, orbiters2.vz)

            da = abs((a-a0)/a0)
            deps = abs(eps-eps0)/eps0

            dev = numpy.where(da > 1.e-12)[0]
            self.assertEqual(len(dev), 0)
            dev = numpy.where(deps > 1.e-12)[0]
            self.assertEqual(len(dev), 0)
            pos[inttype] = [orbiters2.x.value_in(units.AU), orbiters2.y.value_in(units.AU), orbiters2.z.value_in(units.AU)]
        self.assertAlmostEqual(pos[20][0], pos[14][0], 12)
        self.assertAlmostEqual(pos[20][1], pos[14][1], 12)
        self.assertAlmostEqual(pos[20][2], pos[14][2], 12)

    def test25(self):
        print("test massless particles/ kepler integrator, smoothed")
        N = 10
        tend = 20. | units.yr
        numpy.random.seed(12345)
        conv = nbody_system.nbody_to_si(4. | units.MSun, 5. | units.AU)
        orbiters = plummer.new_plummer_model(N, conv)
        sun = datamodel.Particle(mass=1. | units.MSun)
        sun.position = [0, 0, 0] | units.AU
        sun.velocity = [0, 0, 0] | units.kms
        orbiters.mass *= 0.
        eps = (5. | units.AU)

        e0 = energy(sun.mass, eps, orbiters)
        l0 = angular_momentum(orbiters)

        pos = dict()
        for inttype in [20, 14]:
            code = Huayno(conv)
            code.parameters.inttype_parameter = inttype
            code.parameters.timestep_parameter = 0.1
            code.parameters.epsilon_squared = eps**2
            code.particles.add_particle(sun)
            orbiters2 = code.particles.add_particles(orbiters)
            code.evolve_model(tend)

            e1 = energy(sun.mass, eps, orbiters2)
            l1 = angular_momentum(orbiters2)
            de, dl = abs((e1-e0)/e0).max(), abs((l1-l0)/l1).max()
            self.assertTrue(numpy.all(de < 1.e-8))
            self.assertTrue(numpy.all(dl < 1.e-8))

            pos[inttype] = [orbiters2.x.value_in(units.AU), orbiters2.y.value_in(units.AU), orbiters2.z.value_in(units.AU)]
        self.assertAlmostRelativeEqual(pos[20][0], pos[14][0], 4)  # still not clear why 4
        self.assertAlmostRelativeEqual(pos[20][1], pos[14][1], 4)
        self.assertAlmostRelativeEqual(pos[20][2], pos[14][2], 4)

    def test26(self):
        print("test massless particles (negative time)")
        N = 10
        tend = -5. | units.yr
        numpy.random.seed(12345)
        conv = nbody_system.nbody_to_si(4. | units.MSun, 5. | units.AU)
        orbiters = plummer.new_plummer_model(N, conv)
        sun = datamodel.Particle(mass=1. | units.MSun)
        sun.position = [0, 0, 0] | units.AU
        sun.velocity = [0, 0, 0] | units.kms
        orbiters.mass *= 0.

        a0, eps0 = elements(sun.mass, orbiters.x, orbiters.y, orbiters.z,
                      orbiters.vx, orbiters.vy, orbiters.vz)

        pos = dict()
        for inttype in [20, 14]:
            code = Huayno(conv)
            code.parameters.inttype_parameter = inttype
            code.parameters.timestep_parameter = 0.1
            code.particles.add_particle(sun)
            orbiters2 = code.particles.add_particles(orbiters)
            code.evolve_model(tend)
            a, eps = elements(sun.mass, orbiters2.x, orbiters2.y, orbiters2.z,
                      orbiters2.vx, orbiters2.vy, orbiters2.vz)

            da = abs((a-a0)/a0)
            deps = abs(eps-eps0)/eps0

            dev = numpy.where(da > 1.e-12)[0]
            self.assertEqual(len(dev), 0)
            dev = numpy.where(deps > 1.e-12)[0]
            self.assertEqual(len(dev), 0)
            pos[inttype] = [orbiters.x.value_in(units.AU), orbiters.y.value_in(units.AU), orbiters.z.value_in(units.AU)]
        self.assertAlmostEqual(pos[20][0], pos[14][0], 12)
        self.assertAlmostEqual(pos[20][1], pos[14][1], 12)
        self.assertAlmostEqual(pos[20][2], pos[14][2], 12)

    def test27(self):
        particles = plummer.new_plummer_model(31)

        tend = 0.25 | nbody_system.time

        instance = Huayno()
        instance.particles.add_particles(particles)
        instance.evolve_model(tend)
        expected_positions = instance.particles.position
        self.assertEqual(instance.model_time, tend)
        instance.stop()

        particles2 = particles.copy()
        particles2.velocity *= -1

        instance = Huayno()
        instance.particles.add_particles(particles2)
        instance.evolve_model(-tend)
        positions = instance.particles.position
        self.assertEqual(instance.model_time, -tend)
        instance.stop()

        self.assertAlmostEqual(positions, expected_positions)

    def test28(self):
        particles = plummer.new_plummer_model(31)

        instance = Huayno()
        instance.particles.add_particles(particles)
        self.assertAlmostEqual(particles.total_mass(), instance.total_mass)
        self.assertAlmostEqual(particles.center_of_mass(), instance.center_of_mass_position)
        self.assertAlmostEqual(particles.center_of_mass_velocity(), instance.center_of_mass_velocity)

    def test29(self):
        instance = Huayno()
        instance.parameters.epsilon_squared = 0.0 | nbody_system.length**2

        particles = datamodel.Particles(2)
        particles.position = ([0, 0, 0], [1, 0, 0]) | nbody_system.length
        particles.velocity = ([-1, 0, 0], [2, 0, 0]) | nbody_system.speed
        particles.radius = 0 | nbody_system.length
        particles.mass = 0.1 | nbody_system.mass

        instance.particles.add_particles(particles)
        instance.stopping_conditions.out_of_box_detection.enable()
        instance.parameters.stopping_conditions_out_of_box_size = 2 | nbody_system.length
        instance.parameters.stopping_conditions_out_of_box_use_center_of_mass = False
        instance.evolve_model(1 | nbody_system.time)
        self.assertTrue(instance.stopping_conditions.out_of_box_detection.is_set())
        self.assertEqual(len(instance.stopping_conditions.out_of_box_detection.particles(0)), 1)
        self.assertEqual(instance.stopping_conditions.out_of_box_detection.particles(0)[0].key, particles[1].key)
        instance.stop()



import os.path
import numpy
from amuse.test.amusetest import TestWithMPI

from amuse.community.simplex.interface import SimpleXInterface, SimpleX, SimpleXSplitSet
from amuse.units import units
from amuse.datamodel import Particles
from amuse.datamodel import Particle
from amuse.ext.molecular_cloud import ism_cube

# default_options = dict(number_of_workers=2, redirection="none")
default_options = dict(number_of_workers=1)  # ,debugger='gdb')


class TestSimpleXInterface(TestWithMPI):

    def test1(self):
        print("Test 1: initialization")
        instance = SimpleXInterface(**default_options)
        self.assertEqual(0, instance.set_simplex_output_directory(instance.output_directory))
        self.assertEqual(0, instance.initialize_code())
        self.assertEqual(0, instance.commit_parameters())
        self.assertEqual(0, instance.cleanup_code())
        instance.stop()

    def test2(self):
        print("Test 2: commit_particles, getters and setters")
        instance = SimpleXInterface(**default_options)
        self.assertEqual(0, instance.set_simplex_output_directory(instance.output_directory))
        self.assertEqual(0, instance.initialize_code())
        self.assertEqual(0, instance.commit_parameters())

        input_file = os.path.join(os.path.dirname(__file__), "test_simplex_data.txt")
        x, y, z, n_H, flux, X_ion, u = read_input_file(input_file)
        x = numpy.array(x)
        y = numpy.array(y)
        z = numpy.array(z)
        number_of_particles = len(x)
        indices, errors = instance.new_particle(x, y, z, n_H, flux, X_ion, u)
        self.assertEqual(errors, [0]*number_of_particles)
        self.assertEqual(indices, list(range(number_of_particles)))
        self.assertEqual(0, instance.commit_particles())
        x_out, y_out, z_out, n_H_out, flux_out, X_ion_out, u_out, metallicity_out, error = instance.get_state(indices)

        self.assertAlmostEqual((x-x_out)/13200., numpy.zeros_like(x), 7)
        self.assertAlmostEqual((y-y_out)/13200., numpy.zeros_like(x), 7)
        self.assertAlmostEqual((z-z_out)/13200., numpy.zeros_like(x), 7)
        self.assertAlmostEqual(flux, flux_out, 7)
        self.assertAlmostEqual(n_H, n_H_out, 7)
        self.assertAlmostEqual(X_ion, X_ion_out, 7)
        self.assertAlmostRelativeEqual(u, u_out, 7)
        # self.assertAlmostRelativeEqual(metallicity,metallicity_out, 7)

        x, y, z, n_H, flux, X_ion, u, metallicity, error = instance.get_state(0)
        for expected, received in zip([0, 0, 0, 0.001, 5.0, 0.0, 831447247704, 0],
                [x, y, z, n_H, flux, X_ion, u, error]):
            self.assertAlmostRelativeEqual(expected, received, 6)
        x, y, z, error1 = instance.get_position(0)
        n_H, error2 = instance.get_density(0)
        flux, error3 = instance.get_flux(0)
        X_ion, error4 = instance.get_ionisation(0)
        for expected, received in zip([0., 0., 0., 0.001, 5.0, 0.0, 0, 0, 0, 0],
                [x, y, z, n_H, flux, X_ion, error1, error2, error3, error4]):
            self.assertAlmostRelativeEqual(expected, received, 5)

        self.assertEqual(0, instance.set_state(3, 1.0, 2.0, 3.0, 4.0, 5.0, 0.6, 77.0))
        x, y, z, n_H, flux, X_ion, u, metallicity, error = instance.get_state(3)
        for expected, received in zip([1.0, 2.0, 3.0, 4.0, 5.0, 0.6, 77, 0],
                [x, y, z, n_H, flux, X_ion, u, error]):
            self.assertAlmostRelativeEqual(expected, received, 5)
        self.assertEqual(0, instance.set_position(4, 3.0, 2.0, 1.0))
        self.assertEqual(0, instance.set_density(4, 0.6))
        self.assertEqual(0, instance.set_flux(4, 0.5))
        self.assertEqual(0, instance.set_ionisation(4, 0.4))
        self.assertEqual(0, instance.set_internal_energy(4, 1234.))
        x, y, z, n_H, flux, X_ion, u, metallicity, error = instance.get_state(4)
        for expected, received in zip([3.0, 2.0, 1.0, 0.6, 0.5, 0.4, 1234., 0],
                [x, y, z, n_H, flux, X_ion, u, error]):
            self.assertAlmostRelativeEqual(expected, received, 5)

        self.assertEqual(0, instance.set_dinternal_energy_dt(4, 12345.))
        du_dt, err = instance.get_dinternal_energy_dt(4)
        self.assertEqual(12345, du_dt)

        self.assertEqual(0, instance.cleanup_code())
        instance.stop()

    def test3(self):
        print("Test 3: evolve")
        instance = SimpleXInterface(**default_options)
        self.assertEqual(0, instance.set_simplex_output_directory(instance.output_directory))
        self.assertEqual(0, instance.initialize_code())
        self.assertEqual(0, instance.commit_parameters())

        input_file = os.path.join(os.path.dirname(__file__), "test_simplex_data.txt")
        x, y, z, n_H, flux, X_ion, u = read_input_file(input_file)

        number_of_particles = len(x)
        indices, errors = instance.new_particle(x, y, z, n_H, flux, X_ion, u)
        self.assertEqual(errors, [0]*number_of_particles)
        self.assertEqual(indices, list(range(number_of_particles)))

        self.assertEqual(0, instance.commit_particles())

        X_ion, errors = instance.get_ionisation(indices)
        self.assertEqual(errors, [0]*number_of_particles)
        self.assertAlmostEqual(X_ion.sum()/number_of_particles, 0.0)

        density, errors = instance.get_density(indices)
        self.assertEqual(errors, [0]*number_of_particles)
        self.assertAlmostEqual(density.sum(), 1.0, 6)

        self.assertEqual(0, instance.evolve_model(0.5))

        density, errors = instance.get_density(indices)
        self.assertEqual(errors, [0]*number_of_particles)
        self.assertAlmostEqual(density.sum(), 1.0, 6)

        flux, errors = instance.get_flux(indices)
        self.assertEqual(errors, [0]*number_of_particles)
        self.assertEqual(flux.sum(), 5.0)

        X_ion, errors = instance.get_ionisation(indices)
        self.assertEqual(errors, [0]*number_of_particles)
        self.assertAlmostEqual(X_ion.sum()/number_of_particles, 0.000845247683257)

        self.assertEqual(0, instance.cleanup_code())
        instance.stop()

    def test4(self):
        print("Test 4: set boxsize, hilbert_order, timestep")
        instance = SimpleXInterface(**default_options)
        self.assertEqual(0, instance.set_simplex_output_directory(instance.output_directory))
        self.assertEqual(0, instance.initialize_code())

        instance.set_box_size(16384.)
        instance.set_hilbert_order(1)
        instance.set_timestep(0.5)

        self.assertEqual(0, instance.commit_parameters())

        self.assertEqual(16384., instance.get_box_size()['box_size'])
        self.assertEqual(1, instance.get_hilbert_order()['hilbert_order'])
        self.assertEqual(0.5, instance.get_timestep()['timestep'])

        input_file = os.path.join(os.path.dirname(__file__), "test_simplex_data.txt")
        x, y, z, n_H, flux, X_ion, u = read_input_file(input_file)
        number_of_particles = len(x)
        indices, errors = instance.new_particle(x, y, z, n_H, flux, X_ion, u)
        instance.commit_particles()

        self.assertEqual(16384., instance.get_box_size()['box_size'])
        self.assertEqual(1, instance.get_hilbert_order()['hilbert_order'])
        self.assertEqual(0.5, instance.get_timestep()['timestep'])

        self.assertEqual(0, instance.cleanup_code())
        instance.stop()

    def test5(self):
        print("Test 2: delete particles")
        instance = SimpleXInterface(**default_options)
        self.assertEqual(0, instance.set_simplex_output_directory(instance.output_directory))
        self.assertEqual(0, instance.initialize_code())
        self.assertEqual(0, instance.commit_parameters())

        input_file = os.path.join(os.path.dirname(__file__), "test_simplex_data.txt")
        x, y, z, n_H, flux, X_ion, u = read_input_file(input_file)
        x = numpy.array(x)
        y = numpy.array(y)
        z = numpy.array(z)
        number_of_particles = len(x)
        indices, errors = instance.new_particle(x, y, z, n_H, flux, X_ion, u)
        self.assertEqual(errors, [0]*number_of_particles)
        self.assertEqual(indices, list(range(number_of_particles)))
        error = instance.delete_particle(indices[0])
        self.assertEqual(error, -1)
        instance.commit_particles()
        error = instance.delete_particle(indices[0])
# this one I don't understand:
        self.assertEqual(error, -1)
        self.assertEqual(0, instance.evolve_model(0.125))
        error = instance.delete_particle(indices[0])
        self.assertEqual(error, 0)


class TestSimpleX(TestWithMPI):

    def test1(self):
        print("Test 1: initialization")
        instance = SimpleX(**default_options)
        instance.initialize_code()
        instance.commit_parameters()
        instance.cleanup_code()
        instance.stop()

    def test2(self):
        print("Test 2: commit_particles, getters and setters")
        instance = SimpleX(**default_options)
        instance.initialize_code()
        instance.commit_parameters()

        input_file = os.path.join(os.path.dirname(__file__), "test_simplex_data.txt")
        particles = particles_from_input_file(input_file)
        instance.particles.add_particles(particles)
        instance.commit_particles()
#        for attribute in ['position', 'rho', 'flux', 'xion']:
#            self.assertAlmostEqual(13200.*getattr(particles, attribute),
#                                   13200.*getattr(instance.particles, attribute), 5)
#            setattr(instance.particles, attribute, getattr(particles, attribute)/2.0)
#            self.assertAlmostEqual(13200.*getattr(particles, attribute)/2.0,
#                                   13200.*getattr(instance.particles, attribute), 5)
        instance.cleanup_code()
        instance.stop()

    def test3(self):
        print("Test 3: evolve")
        instance = SimpleX(**default_options)
        instance.initialize_code()
        instance.commit_parameters()

        input_file = os.path.join(os.path.dirname(__file__), "test_simplex_data.txt")
        particles = particles_from_input_file(input_file)
        particles.du_dt = particles.u/(10 | units.Myr)
        instance.particles.add_particles(particles)
#        instance.particles.du_dt=particles.du_dt
#        instance.commit_particles()
        instance.particles.du_dt = particles.du_dt
        self.assertAlmostEqual(instance.particles.xion.mean(), 0.0)
        self.assertAlmostEqual(instance.particles.du_dt.mean().in_(units.cm**2/units.s**3), particles.du_dt.mean().in_(units.cm**2/units.s**3))
        instance.evolve_model(0.5 | units.Myr)
        self.assertAlmostEqual(instance.particles.du_dt.mean().in_(units.cm**2/units.s**3), particles.du_dt.mean().in_(units.cm**2/units.s**3))
        self.assertAlmostEqual(instance.particles.xion.mean(), 0.000845247683257)
        instance.cleanup_code()
        instance.stop()

    def test4(self):
        print("Test 4: default parameters")
        instance = SimpleX(**default_options)

        default = dict(timestep=0.05 | units.Myr,
                  source_effective_T=1.e5 | units.K,
                  hilbert_order=1,
                  number_of_freq_bins=1,
                  thermal_evolution_flag=0,
                  blackbody_spectrum_flag=0,
                  box_size=13200 | units.parsec,
                  metal_cooling_flag=0,
                  collisional_ionization_flag=0)
        for x in default:
            self.assertEqual(getattr(instance.parameters, x), default[x])
        instance.commit_parameters()
        for x in default:
            self.assertEqual(getattr(instance.parameters, x), default[x])

        tnow = instance.model_time
        self.assertEqual(tnow, 0. | units.Myr)
        instance.model_time = 321. | units.Myr
        tnow = instance.model_time
        self.assertEqual(tnow, 321. | units.Myr)

    def test5(self):
        print("Test 4: default parameters")
        instance = SimpleX(**default_options)
        param = dict(timestep=0.1 | units.Myr,
                  source_effective_T=2.e5 | units.K,
                  hilbert_order=3,
                  number_of_freq_bins=4,
                  thermal_evolution_flag=1,
                  blackbody_spectrum_flag=1,
                  box_size=32100 | units.parsec,
                  metal_cooling_flag=1,
                  collisional_ionization_flag=1,
                  simplex_data_directory='.')
        for x in param:
            setattr(instance.parameters, x, param[x])
        for x in param:
            self.assertEqual(getattr(instance.parameters, x), param[x])

    def test6(self):
        print("Test 2: print parameters,data directory")
        instance = SimpleX(**default_options)
        print(instance.parameters)

        instance.parameters.simplex_data_directory = "some/dir"
        self.assertEqual(instance.parameters.simplex_data_directory, "some/dir")

    def test7(self):
        print("Test 7: two step evolve")
        instance = SimpleX(**default_options)
        instance.parameters.recombination_radiation_flag = 1

        input_file = os.path.join(os.path.dirname(__file__), "test_simplex_data.txt")
        particles = particles_from_input_file(input_file)
        instance.particles.add_particles(particles)
        self.assertAlmostEqual(instance.particles.xion.mean(), 0.0)
        instance.evolve_model(0.25 | units.Myr)
        flux_before = instance.particles.flux.sum()
        instance.particles.flux = particles.flux
        flux_after = instance.particles.flux.sum()
        self.assertEqual(flux_before, flux_after)
        instance.evolve_model(0.5 | units.Myr)
        self.assertAlmostRelativeEqual(instance.particles.xion.mean(), 0.00084660917243, 3)
        self.assertEqual(instance.particles.flux.max().value_in(1.e48 * units.s**-1), 5)
        instance.cleanup_code()
        instance.stop()

    def test8(self):
        print("Test 8: two step evolve")
        instance = SimpleX(**default_options)

        input_file = os.path.join(os.path.dirname(__file__), "test_simplex_data.txt")
        particles = particles_from_input_file(input_file)
        instance.particles.add_particles(particles)
        self.assertAlmostEqual(instance.particles.xion.mean(), 0.0)
        instance.evolve_model(0.25 | units.Myr)
        instance.evolve_model(0.5 | units.Myr)
        self.assertEqual(instance.particles.flux.max().value_in(1.e48 * units.s**-1), 5)
        self.assertAlmostRelativeEqual(instance.particles.xion.mean(), 0.00084660917243, 3)
        instance.cleanup_code()
        instance.stop()

    def test9(self):
        print("Test 9: add test")
        instance = SimpleX(number_of_workers=1)

        input_file = os.path.join(os.path.dirname(__file__), "test_simplex_data.txt")
        particles = particles_from_input_file(input_file)
        N = len(particles)
        toadd = particles[0:10].copy()
        particles = particles[10:].copy()
        instance.particles.add_particles(particles)
        instance.evolve_model(0.25 | units.Myr)
#        instance.commit_particles()
        instance.particles.add_particles(toadd)
        self.assertEqual(len(instance.particles), N)

    def test10(self):
        print("Test 10: add test")
        instance = SimpleX(number_of_workers=2)

        input_file = os.path.join(os.path.dirname(__file__), "test_simplex_data.txt")
        particles = particles_from_input_file(input_file)
        N = len(particles)
        toadd = particles[0:10].copy()
        particles = particles[10:].copy()
        instance.particles.add_particles(particles)
        instance.evolve_model(0.25 | units.Myr)
#        instance.commit_particles()
        instance.particles.add_particles(toadd)
        self.assertEqual(len(instance.particles), N)

    def test11(self):
        print("Test 11: add test")
        instance = SimpleX(number_of_workers=1)

        input_file = os.path.join(os.path.dirname(__file__), "test_simplex_data.txt")
        particles = particles_from_input_file(input_file)
        N = len(particles)
        toadd = particles[0:10].copy()
        particles = particles[10:].copy()
        instance.particles.add_particles(particles)
        instance.commit_particles()
        instance.particles.add_particles(toadd)
        self.assertEqual(len(instance.particles), N)

    def test12(self):
        N = 1000
        Lstar = 100 | units.LSun
        boxsize = 10 | units.parsec
        rho = 1.0 | (units.amu/units.cm**3)
        t_end = 0.01 | units.Myr
        internal_energy = (9. | units.kms)**2

        source = Particle()
        source.position = (0, 0, 0) | units.parsec
        source.flux = Lstar/(20. | units.eV)
        source.rho = rho
        source.xion = 0.0
        source.u = internal_energy

        ism = ism_cube(N, boxsize/2., rho, internal_energy).result
        ism.rho = rho
        ism.flux = 0. | units.s**-1
        ism.xion = source.xion

        radiative = SimpleX()
        radiative.parameters.box_size = 1.001*boxsize
        radiative.parameters.timestep = 0.001 | units.Myr

        radiative.particles.add_particle(source)
        radiative.particles.add_particles(ism)

        radiative.evolve_model(t_end)
        self.assertAlmostRelativeEquals(0.0750819123073, radiative.particles.xion.mean(), 1)
        radiative.stop()


class TestSimpleXSplitSet(TestWithMPI):

    def test1(self):
        print("Test 1: initialization")
        instance = SimpleXSplitSet(**default_options)
        instance.initialize_code()
        instance.commit_parameters()
        instance.cleanup_code()
        instance.stop()

    def test2(self):
        print("Test 2: commit_particles, getters and setters")
        instance = SimpleXSplitSet(**default_options)
        instance.initialize_code()
        instance.commit_parameters()

        input_file = os.path.join(os.path.dirname(__file__), "test_simplex_data.txt")
        particles, src_particles = splitset_from_input_file(input_file)
        instance.gas_particles.add_particles(particles)
        instance.src_particles.add_particles(src_particles)
        instance.commit_particles()
#        for attribute in ['position', 'rho', 'flux', 'xion']:
#            self.assertAlmostEqual(13200.*getattr(particles, attribute),
#                                   13200.*getattr(instance.particles, attribute), 5)
#            setattr(instance.particles, attribute, getattr(particles, attribute)/2.0)
#            self.assertAlmostEqual(13200.*getattr(particles, attribute)/2.0,
#                                   13200.*getattr(instance.particles, attribute), 5)
        instance.cleanup_code()
        instance.stop()

    def test3(self):
        print("Test 3: evolve")
        instance = SimpleXSplitSet(**default_options)
        instance.initialize_code()
        instance.commit_parameters()

        input_file = os.path.join(os.path.dirname(__file__), "test_simplex_data.txt")
        particles, src_particles = splitset_from_input_file(input_file)
        instance.src_particles.add_particles(src_particles)
        particles.du_dt = particles.u/(10 | units.Myr)
        instance.gas_particles.add_particles(particles)

        self.assertAlmostEqual(instance.gas_particles.xion.mean(), 0.0)
        self.assertAlmostEqual(instance.gas_particles.du_dt.mean().in_(units.cm**2/units.s**3), particles.du_dt.mean().in_(units.cm**2/units.s**3))
        self.assertEqual(instance.get_name_of_current_state(), 'EDIT')
        instance.evolve_model(0.5 | units.Myr)
        self.assertEqual(instance.get_name_of_current_state(), 'RUN')
        self.assertAlmostEqual(instance.gas_particles.du_dt.mean().in_(units.cm**2/units.s**3), particles.du_dt.mean().in_(units.cm**2/units.s**3))
        self.assertAlmostEqual(instance.gas_particles.xion.mean(), 0.000845247683257)
        instance.gas_particles.remove_particles(particles[0:4])
# this is what we would like....
#        self.assertEquals(instance.get_name_of_current_state(), 'UPDATE')
#        instance.recommit_particles()
        instance.evolve_model(0.75 | units.Myr)
        self.assertEqual(len(instance.particles), len(particles)-4)
        instance.cleanup_code()
        instance.stop()

    def test4(self):
        print("Test 4: default parameters")
        instance = SimpleXSplitSet(**default_options)

        default = dict(timestep=0.05 | units.Myr,
                  source_effective_T=1.e5 | units.K,
                  hilbert_order=1,
                  number_of_freq_bins=1,
                  thermal_evolution_flag=0,
                  blackbody_spectrum_flag=0,
                  box_size=13200 | units.parsec,
                  metal_cooling_flag=0,
                  collisional_ionization_flag=0)
        for x in default:
            self.assertEqual(getattr(instance.parameters, x), default[x])
        instance.commit_parameters()
        for x in default:
            self.assertEqual(getattr(instance.parameters, x), default[x])

        tnow = instance.model_time
        self.assertEqual(tnow, 0. | units.Myr)
        instance.model_time = 321. | units.Myr
        tnow = instance.model_time
        self.assertEqual(tnow, 321. | units.Myr)

    def test5(self):
        print("Test 4: default parameters")
        instance = SimpleXSplitSet(**default_options)
        param = dict(timestep=0.1 | units.Myr,
                  source_effective_T=2.e5 | units.K,
                  hilbert_order=3,
                  number_of_freq_bins=4,
                  thermal_evolution_flag=1,
                  blackbody_spectrum_flag=1,
                  box_size=32100 | units.parsec,
                  metal_cooling_flag=1,
                  collisional_ionization_flag=1)
        for x in param:
            setattr(instance.parameters, x, param[x])
        for x in param:
            self.assertEqual(getattr(instance.parameters, x), param[x])


def read_input_file(input_file):
    file = open(input_file, 'r')
    lines = file.readlines()
    lines.pop(0)
    x, y, z, nh, flux, xion, u = [], [], [], [], [], [], []
    for line in lines:
        l = line.strip().split()
        if len(l) >= 7:
            x.append(float(l[1]))
            y.append(float(l[2]))
            z.append(float(l[3]))
            nh.append(float(l[4]))
            flux.append(float(l[5]))
            xion.append(float(l[6]))
            u.append(float(l[7]))

    return x, y, z, nh, flux, xion, u


def particles_from_input_file(input_file):
    x, y, z, n_H, flux, X_ion, u = read_input_file(input_file)
    particles = Particles(len(x))
    particles.x = x | units.parsec
    particles.y = y | units.parsec
    particles.z = z | units.parsec
    particles.rho = n_H | units.amu / units.cm**3
    particles.flux = flux | 1.0e48 / units.s
    particles.xion = X_ion | units.none
    particles.u = u | (units.cm**2/units.s**2)
    return particles


def splitset_from_input_file(input_file):
    x, y, z, n_H, flux, X_ion, u = read_input_file(input_file)

    particles = Particles(len(x))
    particles.x = x | units.parsec
    particles.y = y | units.parsec
    particles.z = z | units.parsec
    particles.rho = n_H | units.amu / units.cm**3
    particles.xion = X_ion | units.none
    particles.u = u | (units.cm**2/units.s**2)

    a = numpy.where(numpy.array(flux) > 0.)[0]
    src_particles = Particles(len(a))
    src_particles.x = numpy.array(x)[a] | units.parsec
    src_particles.y = numpy.array(y)[a] | units.parsec
    src_particles.z = numpy.array(z)[a] | units.parsec
    src_particles.luminosity = numpy.array(flux)[a] | 1.e48*units.s**-1

    return particles, src_particles



import random
import numpy.random
import sys

from amuse.test import amusetest
from amuse.units import units
from amuse.units import constants
from amuse.units import nbody_system
from amuse.support.interface import ConvertArgumentsException

from amuse.ic.plummer import new_plummer_sphere
from amuse.ic.salpeter import new_salpeter_mass_distribution_nbody
from amuse.datamodel import Particle, Particles, ParticlesWithUnitsConverted
from amuse.datamodel import particle_attributes


class TestParticlesAttributes(amusetest.TestCase):

    def test3(self):
        print("Test new_particle_from_cluster_core - nbody units")
        numpy.random.seed(123)
        plummer = new_plummer_sphere(10000)
        result = plummer.new_particle_from_cluster_core(density_weighting_power=1, reuse_hop=True)
        self.assertTrue(isinstance(result, Particle))

        # Casertano & Hut (1985, ApJ, 298, 80):  density weighted core radius = 0.6791 * r_plummer
        plummer_radius = 3 * constants.pi / 16.0 | nbody_system.length
        self.assertAlmostRelativeEqual(result.radius, 0.6791 * plummer_radius, 2)
        self.assertAlmostEqual(result.position, [0.0, 0.0, 0.0] | nbody_system.length, 1)
        self.assertAlmostEqual(result.velocity, [0.0, 0.0, 0.0] | nbody_system.speed, 1)
        self.assertAlmostEqual(result.density, 3.55015420914 | nbody_system.density)

        plummer.vx += 42 | nbody_system.speed
        plummer.vy += (1 - plummer.y / abs(plummer.y).amax()) * (13 | nbody_system.speed)
        plummer.position *= 0.1
        plummer.position += [1.0, 2.0, 3.0] | nbody_system.length
        result = plummer.new_particle_from_cluster_core(density_weighting_power=1, reuse_hop=False)
        self.assertTrue(isinstance(result, Particle))

        self.assertAlmostRelativeEqual(result.radius, 0.06791 * plummer_radius, 2)
        self.assertAlmostEqual(result.position, [1.0, 2.0, 3.0] | nbody_system.length, 1)
        self.assertAlmostEqual(result.velocity, [42.0, 13.0, 0.0] | nbody_system.speed, 1)
        self.assertAlmostRelativeEqual(result.density, 3.55015420914e3 | nbody_system.density, 4)

    def test4(self):
        print("Test new_particle_from_cluster_core - SI units")
        numpy.random.seed(123)
        converter = nbody_system.nbody_to_si(1000.0 | units.MSun, 1.0 | units.parsec)
        plummer = new_plummer_sphere(10000, convert_nbody=converter)
        result = plummer.new_particle_from_cluster_core(unit_converter=converter, density_weighting_power=1, reuse_hop=True)
        self.assertTrue(isinstance(result, Particle))

        # Casertano & Hut (1985, ApJ, 298, 80):  density weighted core radius = 0.6791 * r_plummer
        plummer_radius = 3 * constants.pi / 16.0 | units.parsec
        self.assertAlmostRelativeEqual(result.radius, 0.6791 * plummer_radius, 2)
        self.assertAlmostEqual(result.position, [0.0, 0.0, 0.0] | units.parsec, 1)
        self.assertAlmostEqual(result.velocity, [0.0, 0.0, 0.0] | units.km / units.s, 1)
        self.assertAlmostEqual(result.density, 3.55015420914e3 | units.MSun * units.parsec**-3)

        plummer.vx += 42 | units.km / units.s
        plummer.vy += (1 - plummer.y / abs(plummer.y).amax()) * (13 | units.km / units.s)
        plummer.position *= 0.1
        plummer.position += [1.0, 2.0, 3.0] | units.parsec
        result = plummer.new_particle_from_cluster_core(unit_converter=converter, density_weighting_power=1, reuse_hop=False)
        self.assertTrue(isinstance(result, Particle))

        self.assertAlmostRelativeEqual(result.radius, 0.06791 * plummer_radius, 2)
        self.assertAlmostEqual(result.position, [1.0, 2.0, 3.0] | units.parsec, 1)
        self.assertAlmostEqual(result.velocity, [42.0, 13.0, 0.0] | units.km / units.s, 1)
        self.assertAlmostRelativeEqual(result.density, 3.55015420914e6 | units.MSun * units.parsec**-3, 4)

    def test5(self):
        print("Test new_particle_from_cluster_core - reuse_hop or not")
        converter = nbody_system.nbody_to_si(1.0 | units.MSun, 1.0 | units.parsec)
        plummer = new_plummer_sphere(100, convert_nbody=converter)
        result = plummer.new_particle_from_cluster_core(unit_converter=converter, reuse_hop=True)

        # Hop wasn't stopped, will use same Hop instance:
        result = plummer.new_particle_from_cluster_core(unit_converter=converter, reuse_hop=True)

        nbody_plummer = new_plummer_sphere(100)
        # Hop wasn't stopped, unit_converters don't match:
        self.assertRaises(AttributeError, nbody_plummer.new_particle_from_cluster_core,
            expected_message="Cannot combine units from different systems: m and length")

        result = plummer.new_particle_from_cluster_core(unit_converter=converter, reuse_hop=False)

        # Hop was stopped, new instance will be made with supplied unit_converter (None in this case):
        result = nbody_plummer.new_particle_from_cluster_core(reuse_hop=True)

        self.assertRaises(ConvertArgumentsException, plummer.new_particle_from_cluster_core, unit_converter=converter,  # ,
            expected_message="error while converting parameter 'mass', error: Cannot express kg in mass, the units do not have the same bases")

        result = nbody_plummer.new_particle_from_cluster_core(reuse_hop=False)
        result = plummer.new_particle_from_cluster_core(unit_converter=converter, reuse_hop=False)

    def test6(self):
        print("Test all particle attributes using Hop - each different function creates its own instance of Hop")
        numpy.random.seed(123)
        nbody_plummer = new_plummer_sphere(100)
        nbody_plummer.mass = new_salpeter_mass_distribution_nbody(100)

        # Each different function creates its own instance of Hop
        result = nbody_plummer.new_particle_from_cluster_core(reuse_hop=True)
        result = nbody_plummer.bound_subset(G=nbody_system.G, reuse_hop=True)
        result = nbody_plummer.mass_segregation_Gini_coefficient(reuse_hop=True)
        result = nbody_plummer.LagrangianRadii(reuse_hop=True)
        result = nbody_plummer.densitycentre_coreradius_coredens(reuse_hop=True)

        converter = nbody_system.nbody_to_si(1.0 | units.MSun, 1.0 | units.parsec)
        si_plummer = ParticlesWithUnitsConverted(nbody_plummer, converter.as_converter_from_si_to_nbody())
        functions_using_hop = [
            si_plummer.new_particle_from_cluster_core,
            si_plummer.bound_subset,
            si_plummer.mass_segregation_Gini_coefficient,
            si_plummer.LagrangianRadii,
            si_plummer.densitycentre_coreradius_coredens
        ]

        # Each fails since the Hop instance it tries to reuse has a different unit_converter
        for function_using_hop in functions_using_hop:
            self.assertRaises(Exception, function_using_hop, unit_converter=converter)
        #        expected_message="error while converting parameter 'mass', error: Cannot express kg in mass, the units do not have the same bases (note: check whether Hop needs a converter here)")

        # Close all Hop instances:
        nbody_results = []
        nbody_results.append(nbody_plummer.new_particle_from_cluster_core(reuse_hop=False))
        nbody_results.append(nbody_plummer.bound_subset(G=nbody_system.G, reuse_hop=False))
        nbody_results.append(nbody_plummer.mass_segregation_Gini_coefficient(reuse_hop=False))
        nbody_results.append(nbody_plummer.LagrangianRadii(reuse_hop=False))
        nbody_results.append(nbody_plummer.densitycentre_coreradius_coredens(reuse_hop=False))

        # Now it works, because the Hop instances were closed, and new ones will be instantiated
        si_results = []
        for function_using_hop in functions_using_hop:
            si_results.append(function_using_hop(unit_converter=converter))

        convert = converter.as_converter_from_si_to_nbody()
        self.assertAlmostRelativeEqual(si_results[0].position,
            ParticlesWithUnitsConverted(nbody_results[0].as_set(), convert)[0].position)
        self.assertAlmostRelativeEqual(si_results[1].position,
            ParticlesWithUnitsConverted(nbody_results[1], convert).position)
        self.assertAlmostRelativeEqual(si_results[2], nbody_results[2], places=10)
        self.assertAlmostRelativeEqual(si_results[3][0], convert.from_target_to_source(nbody_results[3][0]), places=10)
        for in_si, in_nbody in zip(si_results[4], nbody_results[4]):
            self.assertAlmostRelativeEqual(in_si, convert.from_target_to_source(in_nbody), places=10)
